"""The CLI for Napistu"""

from __future__ import annotations

import logging
import os
import pickle
from typing import Sequence

import click
import click_logging
import napistu
import igraph as ig
import pandas as pd
from napistu import consensus as napistu_consensus
from napistu import indices
from napistu import sbml_dfs_core
from napistu import utils
from napistu.context import filtering
from napistu.matching import mount
from napistu.ingestion import bigg
from napistu.ingestion import gtex
from napistu.ingestion import hpa
from napistu.ingestion import reactome
from napistu.ingestion import sbml
from napistu.ingestion import string
from napistu.ingestion import trrust
from napistu.modify import curation
from napistu.modify import gaps
from napistu.modify import pathwayannot
from napistu.modify import uncompartmentalize
from napistu.network import net_create
from napistu.network.ig_utils import get_graph_summary
from napistu.network.ng_utils import read_graph_attrs_spec
from napistu.network import precompute
from napistu.ontologies.genodexito import Genodexito
from napistu.ontologies import dogma
from napistu.constants import ONTOLOGIES
from napistu.constants import RESOLVE_MATCHES_AGGREGATORS
from napistu.ingestion.constants import PROTEINATLAS_SUBCELL_LOC_URL
from napistu.ingestion.constants import GTEX_RNASEQ_EXPRESSION_URL
from fs import open_fs

logger = logging.getLogger(napistu.__name__)
click_logging.basic_config(logger)

ALL = "all"


@click.group()
def cli():
    """The Calico Pathway Resources CLI"""
    pass


@click.group()
def ingestion():
    """Command line tools to retrieve raw data."""
    pass


@ingestion.command(name="reactome")
@click.argument("base_folder", type=str)
@click.option(
    "--overwrite", "-o", is_flag=True, default=False, help="Overwrite existing files?"
)
@click_logging.simple_verbosity_option(logger)
def ingest_reactome(base_folder: str, overwrite=True):
    logger.info("Start downloading Reactome to %s", base_folder)
    reactome.reactome_sbml_download(f"{base_folder}/sbml", overwrite=overwrite)


@ingestion.command(name="bigg")
@click.argument("base_folder", type=str)
@click.option(
    "--overwrite", "-o", is_flag=True, default=False, help="Overwrite existing files?"
)
@click_logging.simple_verbosity_option(logger)
def ingest_bigg(base_folder: str, overwrite: bool):
    logger.info("Start downloading Bigg to %s", base_folder)
    bigg.bigg_sbml_download(base_folder, overwrite)


@ingestion.command(name="trrust")
@click.argument("target_uri", type=str)
@click_logging.simple_verbosity_option(logger)
def ingest_ttrust(target_uri: str):
    logger.info("Start downloading TRRUST to %s", target_uri)
    trrust.download_trrust(target_uri)


@ingestion.command(name="proteinatlas-subcell")
@click.argument("target_uri", type=str)
@click.option(
    "--url",
    type=str,
    default=PROTEINATLAS_SUBCELL_LOC_URL,
    help="URL to download the zipped protein atlas subcellular localization tsv from.",
)
@click_logging.simple_verbosity_option(logger)
def ingest_proteinatlas_subcell(target_uri: str, url: str):
    hpa.download_hpa_data(target_uri, url)


@ingestion.command(name="gtex-rnaseq-expression")
@click.argument("target_uri", type=str)
@click.option(
    "--url",
    type=str,
    default=GTEX_RNASEQ_EXPRESSION_URL,
    help="URL to download the gtex file from.",
)
@click_logging.simple_verbosity_option(logger)
def ingest_gtex_rnaseq(target_uri: str, url: str):
    gtex.download_gtex_rnaseq(target_uri, url)


@ingestion.command(name="string-db")
@click.argument("target_uri", type=str)
@click.option(
    "--species",
    type=str,
    default="Homo sapiens",
    help="Species name (e.g., Homo sapiens).",
)
@click_logging.simple_verbosity_option(logger)
def ingest_string_db(target_uri: str, species: str):
    string.download_string(target_uri, species)


@ingestion.command(name="string-aliases")
@click.argument("target_uri", type=str)
@click.option(
    "--species",
    type=str,
    default="Homo sapiens",
    help="Species name (e.g., Homo sapiens).",
)
@click_logging.simple_verbosity_option(logger)
def ingest_string_aliases(target_uri: str, species: str):
    string.download_string_aliases(target_uri, species)


@click.group()
def integrate():
    """Command line tools to integrate raw models into a single SBML_dfs model"""
    pass


@integrate.command(name="reactome")
@click.argument("pw_index_uri", type=str)
@click.argument("output_model_uri", type=str)
@click.option("--species", "-s", multiple=True, default=(ALL,))
@click.option(
    "--overwrite", "-o", is_flag=True, default=False, help="Overwrite existing files?"
)
@click.option(
    "--permissive",
    "-p",
    is_flag=True,
    default=False,
    help="Can parsing failures in submodels throw warnings instead of exceptions?",
)
@click_logging.simple_verbosity_option(logger)
def integrate_reactome(
    pw_index_uri: str,
    output_model_uri: str,
    species: Sequence[str] | None,
    overwrite=False,
    permissive=False,
):
    """Integrates reactome models based on a pw_index"""
    if overwrite is False and utils.path_exists(output_model_uri):
        raise FileExistsError("'output_model_uri' exists but overwrite set False.")
    if species is not None and len(species) == 1 and species[0] == ALL:
        species = None

    strict = not permissive
    logger.debug(f"permissive = {permissive}; strict = {strict}")

    consensus_model = reactome.construct_reactome_consensus(
        pw_index_uri, species=species, strict=strict
    )
    utils.save_pickle(output_model_uri, consensus_model)


@integrate.command(name="bigg")
@click.argument("pw_index_uri", type=str)
@click.argument("output_model_uri", type=str)
@click.option("--species", "-s", multiple=True, default=(ALL,))
@click.option(
    "--overwrite", "-o", is_flag=True, default=False, help="Overwrite existing files?"
)
@click_logging.simple_verbosity_option(logger)
def integrate_bigg(
    pw_index_uri: str,
    output_model_uri: str,
    species: Sequence[str] | None,
    overwrite=False,
):
    """Integrates bigg models based on a pw_index"""
    if overwrite is False and utils.path_exists(output_model_uri):
        raise FileExistsError("'output_model_uri' exists but overwrite set False.")
    if species is not None and len(species) == 1 and species[0] == ALL:
        species = None
    consensus_model = bigg.construct_bigg_consensus(pw_index_uri, species=species)
    utils.save_pickle(output_model_uri, consensus_model)


@integrate.command(name="trrust")
@click.argument("trrust_csv_uri", type=str)
@click.argument("output_model_uri", type=str)
@click.option(
    "--overwrite", "-o", is_flag=True, default=False, help="Overwrite existing files?"
)
@click_logging.simple_verbosity_option(logger)
def integrate_trrust(
    trrust_csv_uri: str,
    output_model_uri: str,
    overwrite=False,
):
    """Converts TRRUST csv to SBML_dfs model"""
    if overwrite is False and utils.path_exists(output_model_uri):
        raise FileExistsError("'output_model_uri' exists but overwrite set False.")
    logger.info("Start converting TRRUST csv to SBML_dfs")
    sbmldfs_model = trrust.convert_trrust_to_sbml_dfs(trrust_csv_uri)
    logger.info("Save SBML_dfs model to %s", output_model_uri)
    utils.save_pickle(output_model_uri, sbmldfs_model)


@integrate.command(name="string-db")
@click.argument("string_db_uri", type=str)
@click.argument("string_aliases_uri", type=str)
@click.argument("output_model_uri", type=str)
@click.option(
    "--overwrite", "-o", is_flag=True, default=False, help="Overwrite existing files?"
)
@click_logging.simple_verbosity_option(logger)
def integrate_string_db(
    string_db_uri: str, string_aliases_uri: str, output_model_uri: str, overwrite=False
):
    """Converts string-db to the sbml_dfs format"""
    if overwrite is False and utils.path_exists(output_model_uri):
        raise FileExistsError("'output_model_uri' exists but overwrite set False.")
    logger.info("Start converting string-db to SBML_dfs")
    sbmldfs_model = string.convert_string_to_sbml_dfs(string_db_uri, string_aliases_uri)
    logger.info("Save SBML_dfs model to %s", output_model_uri)
    utils.save_pickle(output_model_uri, sbmldfs_model)


@click.group()
def consensus():
    """Command line tools to create a consensus model from SBML_dfs"""
    pass


@consensus.command(name="create")
@click.argument("sbml_dfs_uris", type=str, nargs=-1)
@click.argument("output_model_uri", type=str, nargs=1)
@click.option(
    "--nondogmatic",
    "-n",
    is_flag=True,
    default=False,
    help="Run in non-dogmatic mode (trying to merge genes and proteins)?",
)
@click_logging.simple_verbosity_option(logger)
def create_consensus(
    sbml_dfs_uris: Sequence[str], output_model_uri: str, nondogmatic: bool
):
    """Create a consensus model from a list of SBML_dfs"""

    dogmatic = not nondogmatic
    logger.debug(f"nondogmatic = {nondogmatic}; dogmatic = {dogmatic}")
    logger.info(
        f"Creating a consensus from {len(sbml_dfs_uris)} sbml_dfs where dogmatic = {dogmatic}"
    )

    sbml_dfs_dict = {uri: utils.load_pickle(uri) for uri in sbml_dfs_uris}
    pw_index_df = pd.DataFrame(
        {
            "file": sbml_dfs_uris,
            "pathway_id": sbml_dfs_dict.keys(),
            "source": sbml_dfs_dict.keys(),
            "name": sbml_dfs_dict.keys(),
            # TODO: Discuss with Sean how to deal with date in pw_index
            "date": "1900-01-01",
        }
    )
    pw_index_df["species"] = "unknown"
    pw_index = indices.PWIndex(pw_index=pw_index_df, validate_paths=False)
    consensus_model = napistu_consensus.construct_consensus_model(
        sbml_dfs_dict, pw_index, dogmatic
    )
    utils.save_pickle(output_model_uri, consensus_model)


@click.group()
def refine():
    """Command line tools to refine a consensus model"""
    pass


@refine.command(name="add_reactome_entity_sets")
@click.argument("model_uri", type=str)
@click.argument("entity_set_csv", type=str)
@click.argument("output_model_uri", type=str)
def add_reactome_entity_sets(
    model_uri: str, entity_set_csv: str, output_model_uri: str
):
    """Add reactome entity sets to a consensus model

    The entity set csv is classically exported from the neo4j reactome
    database.
    """
    model = utils.load_pickle(model_uri)
    model = pathwayannot.add_entity_sets(model, entity_set_csv)
    utils.save_pickle(output_model_uri, model)


@refine.command(name="add_reactome_identifiers")
@click.argument("model_uri", type=str)
@click.argument("crossref_csv", type=str)
@click.argument("output_model_uri", type=str)
def add_reactome_identifiers(model_uri: str, crossref_csv: str, output_model_uri: str):
    """Add reactome identifiers to a consensus model

    The crossref csv is classically exported from the neo4j reactome
    database.
    """
    model = utils.load_pickle(model_uri)
    model = pathwayannot.add_reactome_identifiers(model, crossref_csv)
    utils.save_pickle(output_model_uri, model)


@refine.command(name="infer_uncompartmentalized_species_location")
@click.argument("model_uri", type=str)
@click.argument("output_model_uri", type=str)
def infer_uncompartmentalized_species_location(model_uri: str, output_model_uri: str):
    """
    Infer Uncompartmentalized Species Location

    If the compartment of a subset of compartmentalized species was
    not specified, infer an appropriate compartment from other members of reactions they particpate in
    """
    sbml_dfs = utils.load_pickle(model_uri)
    sbml_dfs.infer_uncompartmentalized_species_location()
    utils.save_pickle(output_model_uri, sbml_dfs)


@refine.command(name="name_compartmentalized_species")
@click.argument("model_uri", type=str)
@click.argument("output_model_uri", type=str)
def name_compartmentalized_species(model_uri: str, output_model_uri: str):
    """
    Name Compartmentalized Species

    Rename compartmentalized species if they have the same name as their species
    """
    sbml_dfs = utils.load_pickle(model_uri)
    sbml_dfs.name_compartmentalized_species()
    utils.save_pickle(output_model_uri, sbml_dfs)


@refine.command(name="merge_model_compartments")
@click.argument("model_uri", type=str)
@click.argument("output_model_uri", type=str)
def merge_model_compartments(model_uri: str, output_model_uri: str):
    """Take a compartmentalized mechanistic model and merge all of the compartments."""
    model = utils.load_pickle(model_uri)
    model = uncompartmentalize.uncompartmentalize_sbml_dfs(model)
    utils.save_pickle(output_model_uri, model)


@refine.command(name="drop_cofactors")
@click.argument("model_uri", type=str)
@click.argument("output_model_uri", type=str)
def drop_cofactors(model_uri: str, output_model_uri: str):
    """Remove reaction species acting as cofactors"""
    model = utils.load_pickle(model_uri)
    model = pathwayannot.drop_cofactors(model)
    utils.save_pickle(output_model_uri, model)


@refine.command(name="add_transportation_reactions")
@click.argument("model_uri", type=str)
@click.argument("output_model_uri", type=str)
@click.option(
    "--exchange-compartment",
    "-e",
    default="cytosol",
    help="Exchange compartment for new transport reactions.",
)
@click_logging.simple_verbosity_option(logger)
def add_transportation_reaction(
    model_uri, output_model_uri, exchange_compartment="cytosol"
):
    """Add transportation reactions to a consensus model"""

    model = utils.load_pickle(model_uri)
    model = gaps.add_transportation_reactions(
        model, exchange_compartment=exchange_compartment
    )
    utils.save_pickle(output_model_uri, model)


@refine.command(name="apply_manual_curations")
@click.argument("model_uri", type=str)
@click.argument("curation_dir", type=str)
@click.argument("output_model_uri", type=str)
def apply_manual_curations(model_uri: str, curation_dir: str, output_model_uri: str):
    """Apply manual curations to a consensus model

    The curation dir is a directory containing the manual curations
    Check napistu.modify.curation.curate_sbml_dfs for more information.
    """
    model = utils.load_pickle(model_uri)
    model = curation.curate_sbml_dfs(curation_dir=curation_dir, sbml_dfs=model)
    utils.save_pickle(output_model_uri, model)


@refine.command(name="expand_identifiers")
@click.argument("sbml_dfs_uri", type=str)
@click.argument("output_model_uri", type=str)
@click.option("--species", "-s", default="Homo sapiens", type=str)
@click.option(
    "--ontologies", "-o", multiple=True, type=str, help="Ontologies to add or complete"
)
@click.option(
    "--preferred_method",
    "-p",
    default="bioconductor",
    type=str,
    help="Preferred method to use for identifier expansion",
)
@click.option(
    "--allow_fallback",
    "-a",
    default=True,
    type=bool,
    help="Allow fallback to other methods if preferred method fails",
)
def expand_identifiers(
    sbml_dfs_uri: str,
    output_model_uri: str,
    species: str,
    ontologies: set[str],
    preferred_method: str,
    allow_fallback: bool,
):
    """Expand identifiers of a model

    Args:
        sbml_dfs_uri (str): uri of model in sbml dfs format
        output_model_uri (str): output uri of model in sbml dfs format
        species (str): Species to use
        ontologies (set[str]): ontologies to add or update

    Example call:
    > cpr refine expand_identifiers gs://<uri> ./test.pickle -o ensembl_gene
    """
    sbml_dfs: sbml.SBML_dfs = utils.load_pickle(sbml_dfs_uri)  # type: ignore
    if len(ontologies) == 0:
        raise ValueError("No ontologies to expand specified.")

    Genodexito(
        species=species,
        preferred_method=preferred_method,
        allow_fallback=allow_fallback,
    ).expand_sbml_dfs_ids(sbml_dfs, ontologies=ontologies)

    utils.save_pickle(output_model_uri, sbml_dfs)


@integrate.command(name="dogmatic_scaffold")
@click.argument("output_model_uri", type=str)
@click.option("--species", "-s", default="Homo sapiens", type=str)
@click.option(
    "--preferred_method",
    "-p",
    default="bioconductor",
    type=str,
    help="Preferred method to use for identifier expansion",
)
@click.option(
    "--allow_fallback",
    "-a",
    default=True,
    type=bool,
    help="Allow fallback to other methods if preferred method fails",
)
def dogmatic_scaffold(
    output_model_uri: str,
    species: str,
    preferred_method: str,
    allow_fallback: bool,
):
    """Dogmatic Scaffold

    Args:
        output_model_uri (str): output uri of model in sbml dfs format
        species (str): Species to use

    Example call:
    > cpr integrate dogmatic_scaffold ./test.pickle
    """

    dogmatic_sbml_dfs = dogma.create_dogmatic_sbml_dfs(
        species=species,
        preferred_method=preferred_method,
        allow_fallback=allow_fallback,
    )

    utils.save_pickle(output_model_uri, dogmatic_sbml_dfs)


@refine.command(name="filter_gtex_tissue")
@click.argument("sbml_dfs_uri", type=str)
@click.argument("gtex_file_uri", type=str)
@click.argument("output_model_uri", type=str)
@click.argument("tissue", type=str)
@click_logging.simple_verbosity_option(logger)
def filter_gtex_tissue(
    sbml_dfs_uri: str,
    gtex_file_uri: str,
    output_model_uri: str,
    tissue: str,
    filter_non_genic_reactions: bool,
):
    """Filter model by the gtex tissue expression

    This uses zfpkm values derived from gtex to filter the model.
    """

    logger.info("Load sbml_dfs model")
    sbml_dfs: sbml.SBML_dfs = utils.load_pickle(sbml_dfs_uri)  # type: ignore
    logger.info("Load and clean gtex tissue expression")
    dat_gtex = gtex.load_and_clean_gtex_data(gtex_file_uri)
    logger.info("Annotate genes with gtex tissue expression")
    mount.bind_wide_results(
        sbml_dfs=sbml_dfs,
        results_df=dat_gtex.reset_index(drop=False),
        results_name="gtex",
        ontologies={ONTOLOGIES.ENSEMBL_GENE},
        numeric_agg=RESOLVE_MATCHES_AGGREGATORS.MAX,
    )
    logger.info("Trim network by gene attribute")
    filtering.filter_species_by_attribute(
        sbml_dfs,
        "gtex",
        attribute_name=tissue,
        # remove entries which are NOT in the liver
        attribute_value=0,
        inplace=True,
    )
    # remove the gtex species data from the sbml_dfs
    sbml_dfs.remove_species_data("gtex")

    logger.info("Save sbml_dfs to %s", output_model_uri)
    utils.save_pickle(output_model_uri, sbml_dfs)


@refine.command(name="filter_hpa_compartments")
@click.argument("sbml_dfs_uri", type=str)
@click.argument("hpa_file_uri", type=str)
@click.argument("output_model_uri", type=str)
@click_logging.simple_verbosity_option(logger)
def filter_hpa_gene_compartments(
    sbml_dfs_uri: str, hpa_file_uri: str, output_model_uri: str
):
    """Filter an interaction network using the human protein atlas

    This uses loads the human proteome atlas and removes reactions (including interactions)
    containing genes which are not colocalized.

    Only interactions between genes in the same compartment are kept.
    """

    logger.info("Load sbml_dfs model")
    sbml_dfs: sbml.SBML_dfs = utils.load_pickle(sbml_dfs_uri)  # type: ignore
    logger.info("Load and clean hpa data")
    dat_hpa = hpa.load_and_clean_hpa_data(hpa_file_uri)
    logger.info("Annotate genes with HPA compartments")
    mount.bind_wide_results(
        sbml_dfs=sbml_dfs,
        results_df=dat_hpa.reset_index(drop=False),
        results_name="hpa",
        ontologies={ONTOLOGIES.ENSEMBL_GENE},
        numeric_agg=RESOLVE_MATCHES_AGGREGATORS.MAX,
    )
    logger.info(
        "Trim network removing reactions with species in different compartments"
    )
    filtering.filter_reactions_with_disconnected_cspecies(
        sbml_dfs, "hpa", inplace=False
    )
    sbml_dfs.remove_species_data("hpa")

    logger.info("Save sbml_dfs to %s", output_model_uri)
    utils.save_pickle(output_model_uri, sbml_dfs)


@click.group()
def exporter():
    """Command line tools to export a consensus model
    to various formats
    """
    pass


@exporter.command(name="export_igraph")
@click.argument("model_uri", type=str)
@click.argument("output_uri", type=str)
@click.option(
    "--graph_attrs_spec_uri",
    "-a",
    default=None,
    help="File specifying reaction and/or species attributes to add to the graph",
)
@click.option(
    "--format", "-f", default="pickle", help="Output format: gml, edgelist, pickle"
)
@click.option(
    "--wiring_approach",
    "-g",
    type=str,
    default="bipartite",
    help="bipartite or regulatory",
)
@click.option(
    "--weighting_strategy",
    "-w",
    type=str,
    default="unweighted",
    help="Approach to adding weights to the network",
)
@click.option(
    "--directed", "-d", type=bool, default=True, help="Directed or undirected graph?"
)
@click.option(
    "--reverse",
    "-r",
    type=bool,
    default=False,
    help="Reverse edges so they flow from effects to causes?",
)
def export_igraph(
    model_uri: str,
    output_uri: str,
    graph_attrs_spec_uri: str | None,
    format: str,
    wiring_approach: str,
    weighting_strategy: str,
    directed: bool,
    reverse: bool,
):
    """Export the consensus model as an igraph object"""
    model = utils.load_pickle(model_uri)

    if graph_attrs_spec_uri is None:
        graph_attrs_spec = None
    else:
        graph_attrs_spec = read_graph_attrs_spec(graph_attrs_spec_uri)

    napistu_graph = net_create.process_napistu_graph(
        model,
        reaction_graph_attrs=graph_attrs_spec,
        directed=directed,
        edge_reversed=reverse,
        wiring_approach=wiring_approach,
        weighting_strategy=weighting_strategy,
        verbose=True,
    )

    base, path = os.path.split(output_uri)
    with open_fs(base, create=True, writeable=True) as fs:
        with fs.openbin(path, "wb") as f:
            if format == "gml":
                napistu_graph.write_gml(f)
            elif format == "edgelist":
                napistu_graph.write_edgelist(f)
            elif format == "pickle":
                pickle.dump(napistu_graph, f)
            else:
                raise ValueError("Unknown format: %s" % format)


@exporter.command(name="export_precomputed_distances")
@click.argument("graph_uri", type=str)
@click.argument("output_uri", type=str)
@click.option(
    "--format",
    "-f",
    type=str,
    default="pickle",
    help="Input igraph format: gml, edgelist, pickle",
)
@click.option(
    "--max_steps",
    "-s",
    type=int,
    default=-1,
    help="The max number of steps between pairs of species to save a distance",
)
@click.option(
    "--max_score_q",
    "-q",
    type=float,
    default=1,
    help='Retain up to the "max_score_q" quantiles of all scores (small scores are better)',
)
@click.option(
    "--partition_size",
    "-p",
    type=int,
    default=5000,
    help="The number of species to process together when computing distances",
)
@click.option(
    "--weights_vars",
    "-w",
    type=str,
    default=["weights", "upstream_weights"],
    help="One or more variables defining edge weights to use when calculating weighted shortest paths.",
)
def export_precomputed_distances(
    graph_uri: str,
    output_uri: str,
    format: str,
    max_steps: int,
    max_score_q: float,
    partition_size: int,
    weights_vars: str,
):
    """Export precomputed distances for the igraph object"""

    base, path = os.path.split(graph_uri)
    with open_fs(base) as fs:
        with fs.openbin(path) as f:
            if format == "gml":
                napistu_graph = ig.Graph.Read_GML(f)
            elif format == "edgelist":
                napistu_graph = ig.Graph.Read_Edgelist(f)
            elif format == "pickle":
                napistu_graph = ig.Graph.Read_Pickle(f)
            else:
                raise ValueError("Unknown format: %s" % format)

    # convert weight vars from a str to list
    weights_vars_list = utils.click_str_to_list(weights_vars)

    precomputed_distances = precompute.precompute_distances(
        napistu_graph,
        max_steps=max_steps,
        max_score_q=max_score_q,
        partition_size=partition_size,
        weights_vars=weights_vars_list,
    )

    utils.save_parquet(precomputed_distances, output_uri)


@exporter.command(name="export_smbl_dfs_tables")
@click.argument("model_uri", type=str)
@click.argument("output_uri", type=str)
@click.option(
    "--overwrite", "-o", is_flag=True, default=False, help="Overwrite existing files?"
)
@click.option(
    "--model-prefix", "-m", type=str, default="", help="Model prefix for files?"
)
@click.option(
    "--nondogmatic",
    "-n",
    is_flag=True,
    default=False,
    help="Run in non-dogmatic mode (trying to merge genes and proteins)?",
)
@click_logging.simple_verbosity_option(logger)
def export_sbml_dfs_tables(
    model_uri: str,
    output_uri: str,
    overwrite=False,
    model_prefix="",
    nondogmatic: bool = True,
):
    """Export the consensus model as a collection of table"""

    dogmatic = not nondogmatic
    logger.debug(f"nondogmatic = {nondogmatic}; dogmatic = {dogmatic}")
    logger.info(f"Exporting tables with dogmatic = {dogmatic}")

    sbml_dfs = utils.load_pickle(model_uri)
    sbml_dfs.export_sbml_dfs(
        model_prefix, output_uri, overwrite=overwrite, dogmatic=dogmatic
    )


@click.group()
def importer():
    """Tools to import sbml_dfs directly form other sources"""
    pass


@importer.command(name="sbml_dfs")
@click.argument("input_uri", type=str)
@click.argument("output_uri", type=str)
@click_logging.simple_verbosity_option(logger)
def import_sbml_dfs_from_sbml_dfs_uri(input_uri, output_uri):
    """Import sbml_dfs from an uri, eg another GCS bucket"""
    logger.info("Load sbml_dfs from %s", input_uri)
    # We could also just copy the file, but I think validating
    # the filetype is a good idea to prevent downstream errors.
    sbml_dfs = utils.load_pickle(input_uri)
    if not (isinstance(sbml_dfs, sbml.SBML_dfs)):
        raise ValueError(
            f"Pickled input is not an SBML_dfs object but {type(sbml_dfs)}: {input_uri}"
        )
    logger.info("Save file to %s", output_uri)
    utils.save_pickle(output_uri, sbml_dfs)


@importer.command(name="sbml")
@click.argument("input_uri", type=str)
@click.argument("output_uri", type=str)
@click_logging.simple_verbosity_option(logger)
def import_sbml_dfs_from_sbml(input_uri, output_uri):
    """Import sbml_dfs from a sbml file"""
    logger.info("Load sbml from %s", input_uri)
    # We could also just copy the file, but I think validating
    # the filetype is a good idea to prevent downstream errors.
    sbml_file = sbml.SBML(input_uri)
    logger.info("Convert file to sbml_dfs")
    sbml_dfs = sbml_dfs_core.SBML_dfs(sbml_file)
    logger.info("Save file to %s", output_uri)
    utils.save_pickle(output_uri, sbml_dfs)


@click.group()
def contextualizer():
    """Command line tools to contextualize a pathway model"""
    pass


@click.group()
def helpers():
    """Various helper functions"""
    pass


@helpers.command(name="copy_uri")
@click.argument("input_uri", type=str)
@click.argument("output_uri", type=str)
@click.option("--is-file", type=bool, default=True, help="Is the input a file?")
@click_logging.simple_verbosity_option(logger)
def copy_uri(input_uri, output_uri, is_file=True):
    """Copy a uri representing a file or folder from one location to another"""
    logger.info("Copy uri from %s to %s", input_uri, output_uri)
    utils.copy_uri(input_uri, output_uri, is_file=is_file)


@helpers.command(name="validate_sbml_dfs")
@click.argument("input_uri", type=str)
@click_logging.simple_verbosity_option(logger)
def validate_sbml_dfs(input_uri):
    """Validate a sbml_dfs object"""
    sbml_dfs = utils.load_pickle(input_uri)
    sbml_dfs.validate()

    logger.info(f"Successfully validated: {input_uri}")


@click.group()
def stats():
    """Various functions to calculate network statistics

    The statistics are saved as json files
    """
    pass


@stats.command(name="sbml_dfs_network")
@click.argument("input_uri", type=str)
@click.argument("output_uri", type=str)
def calculate_sbml_dfs_stats(input_uri, output_uri):
    """Calculate statistics for a sbml_dfs object"""
    model: sbml_dfs_core.SBML_dfs = utils.load_pickle(input_uri)  # type: ignore
    stats = model.get_network_summary()
    utils.save_json(output_uri, stats)


@stats.command(name="igraph_network")
@click.argument("input_uri", type=str)
@click.argument("output_uri", type=str)
def calculate_igraph_stats(input_uri, output_uri):
    """Calculate statistics for an igraph object"""
    graph: ig.Graph = utils.load_pickle(input_uri)  # type: ignore
    stats = get_graph_summary(graph)
    utils.save_json(output_uri, stats)


cli.add_command(ingestion)
cli.add_command(integrate)
cli.add_command(consensus)
cli.add_command(refine)
cli.add_command(exporter)
cli.add_command(importer)
cli.add_command(helpers)
cli.add_command(stats)

if __name__ == "__main__":
    cli()



import pandas as pd

from napistu.network import paths
from napistu.network import ng_utils


def test_shortest_paths(sbml_dfs, napistu_graph, napistu_graph_undirected):
    species = sbml_dfs.species
    source_species = species[species["s_name"] == "NADH"]
    dest_species = species[species["s_name"] == "NAD+"]
    target_species_paths = ng_utils.compartmentalize_species_pairs(
        sbml_dfs, source_species.index.tolist(), dest_species.index.tolist()
    )

    (
        all_shortest_reaction_paths_df,
        _,
        _,
        _,
    ) = paths.find_all_shortest_reaction_paths(
        napistu_graph, sbml_dfs, target_species_paths, weight_var="weights"
    )

    # undirected graph
    (
        all_shortest_reaction_paths_df,
        all_shortest_reaction_path_edges_df,
        edge_sources,
        paths_graph,
    ) = paths.find_all_shortest_reaction_paths(
        napistu_graph_undirected, sbml_dfs, target_species_paths, weight_var="weights"
    )

    assert all_shortest_reaction_paths_df.shape[0] == 3


def test_net_polarity():
    polarity_series = pd.Series(
        ["ambiguous", "ambiguous"], index=[0, 1], name="link_polarity"
    )
    assert all(
        [x == "ambiguous" for x in paths._calculate_net_polarity(polarity_series)]
    )

    polarity_series = pd.Series(
        ["activation", "inhibition", "inhibition", "ambiguous"],
        index=range(0, 4),
        name="link_polarity",
    )
    assert paths._calculate_net_polarity(polarity_series) == [
        "activation",
        "inhibition",
        "activation",
        "ambiguous activation",
    ]
    assert paths._terminal_net_polarity(polarity_series) == "ambiguous activation"



"""Module to contain constants for the modify submodule"""

from __future__ import annotations

from types import SimpleNamespace

import pandas as pd

from napistu.constants import IDENTIFIERS
from napistu.constants import ONTOLOGIES

VALID_ANNOTATION_TYPES = [
    "foci",
    "reactions",
    "species",
    "compartments",
    "compartmentalized_species",
    "reaction_species",
    "remove",
]

# if_all defines reactions species which must all be present for a filter to occur
# except_any defines reaction species which will override "if_all"
# as_substrates defines reaction species which must be present as a substrate for filtering to occur
COFACTOR_SCHEMA = {
    "ATP PO4 donation": {"if_all": ["ATP", "ADP"], "except_any": ["AMP"]},
    "GTP PO4 donation": {"if_all": ["GTP", "GDP"]},
    "ATP PPi donation": {"if_all": ["ATP", "AMP"], "except_any": ["ADP"]},
    "NADH H- donation": {"if_all": ["NADH", "NAD+"], "as_substrate": ["NADH"]},
    "NADPH H- donation": {"if_all": ["NADPH", "NADP+"], "as_substrate": ["NADPH"]},
    "SAH methyltransferase": {"if_all": ["SAH", "SAM"]},
    "Glutathione oxidation": {"if_all": ["GSSG", "GSH"], "except_any": ["NADPH"]},
    # "Glutamine aminotransferase" :
    #    {"if_all" : ["Gln", "Glu"],
    #     "except_any" : ["ATP"]},
    "Water": {"if_all": ["water"]},
    "PO4": {"if_all": ["PO4"]},
    "PPi": {"if_all": ["PPi"]},
    "H+": {"if_all": ["H+"]},
    "O2": {"if_all": ["O2"]},
    "CO2": {"if_all": ["CO2"]},
    "Na+": {"if_all": ["Na+"]},
    "Cl-": {"if_all": ["Cl-"]},
    "CoA": {"if_all": ["CoA"]},
    "HCO3-": {"if_all": ["HCO3"]},
}

COFACTOR_CHEBI_IDS = pd.DataFrame(
    [
        ("ADP", 456216),  # ADP(3−)
        ("ADP", 16761),
        ("AMP", 16027),
        ("ATP", 30616),  # ATP(4-)
        ("ATP", 15422),
        ("CO2", 16526),
        ("HCO3", 17544),
        ("H2CO3", 28976),
        ("GDP", 17552),
        ("GSH", 16856),
        ("GSSG", 17858),
        ("GTP", 15996),
        ("Glu", 29985),
        ("Gln", 58359),
        ("H+", 15378),
        ("H+", 24636),
        ("O2", 15379),
        ("NADH", 57945),  # NADH(2−)
        ("NADH", 16908),  # NADH
        ("NAD+", 57540),  # NAD(1-)
        ("NAD+", 15846),  # NAD(+)
        ("NADPH", 16474),
        ("NADP+", 18009),
        ("NADP+", 58349),  # NADP(3−)
        ("PO4", 18367),
        ("PPi", 29888),  # H2PO4
        ("PPi", 18361),  # PPi4-
        ("SAH", 16680),
        ("SAM", 15414),
        ("water", 15377),
        ("water", 16234),  # HO-
        ("Na+", 29101),
        ("Cl-", 29311),
        ("CoA", 1146900),
        ("CoA", 57287),
        ("acetyl-CoA", 15351),
        ("FAD", 16238),
        ("FADH2", 17877),
        ("UDP", 17659),
    ],
    columns=["cofactor", "chebi"],
)

NEO4J_MEMBERS_RAW = SimpleNamespace(
    SET_NAME="set_name",
    SET_ID="set_id",
    MEMBER_NAME="member_name",
    MEMBER_ID="member_id",
    IDENTIFIER=IDENTIFIERS.IDENTIFIER,
    ONTOLOGY=IDENTIFIERS.ONTOLOGY,
)

NEO4_MEMBERS_SET = {
    NEO4J_MEMBERS_RAW.SET_NAME,
    NEO4J_MEMBERS_RAW.SET_ID,
    NEO4J_MEMBERS_RAW.MEMBER_NAME,
    NEO4J_MEMBERS_RAW.MEMBER_ID,
    NEO4J_MEMBERS_RAW.IDENTIFIER,
    NEO4J_MEMBERS_RAW.ONTOLOGY,
}

REACTOME_CROSSREF_RAW = SimpleNamespace(
    MEMBER_NAME="member_name",
    REACTOME_ID="reactome_id",
    UNIPROT=ONTOLOGIES.UNIPROT,
    IDENTIFIER=IDENTIFIERS.IDENTIFIER,
    ONTOLOGY=IDENTIFIERS.ONTOLOGY,
    URL=IDENTIFIERS.URL,
)

REACTOME_CROSSREF_SET = {
    REACTOME_CROSSREF_RAW.MEMBER_NAME,
    REACTOME_CROSSREF_RAW.REACTOME_ID,
    REACTOME_CROSSREF_RAW.UNIPROT,
    REACTOME_CROSSREF_RAW.IDENTIFIER,
    REACTOME_CROSSREF_RAW.ONTOLOGY,
    REACTOME_CROSSREF_RAW.URL,
}



from __future__ import annotations

import pandas as pd
from napistu import identifiers
from napistu import sbml_dfs_core
from napistu import sbml_dfs_utils
from napistu import source
from napistu import utils
from napistu.constants import BQB
from napistu.constants import MINI_SBO_FROM_NAME
from napistu.ingestion.constants import YEAST_IDEA_KINETICS_URL
from napistu.ingestion.constants import YEAST_IDEA_PUBMED_ID
from napistu.ingestion.constants import YEAST_IDEA_SOURCE
from napistu.ingestion.constants import YEAST_IDEA_TARGET


def download_idea(output_dir: str) -> None:
    # save to
    utils.download_and_extract(YEAST_IDEA_KINETICS_URL, output_dir)

    # TODO: since only a single file is outputted, it makes sense to download and extract the data, then copy it to a target URI
    # TODO: add GCS support
    pass


def convert_idea_kinetics_to_sbml_dfs(
    idea_path: str,
) -> sbml_dfs_core.SBML_dfs:
    """
    Convert IDEA Kinetics to SBML DFs

    Format yeast induction regulator->target relationships as a directed graph.

    Args:
        idea_path: Path to the IDEA Kinetics file.

    Returns:
        SBML_dfs: an SBML_dfs object containing molecular species and their interactions.
        Kinetic attributes are included as reactions_data.

    """

    # TO DO - replace with GCS support (currently this just reads a local .tsv)
    idea_kinetics_df = pd.read_csv(idea_path, sep="\t")

    # separate based on whether the change is probably direct or indirect
    idea_kinetics_df["directness"] = [
        "direct" if t_rise < 15 else "indirect" for t_rise in idea_kinetics_df["t_rise"]
    ]

    # reduce cases of multiple TF-target pairs to a single entry
    distinct_edges = (
        idea_kinetics_df.groupby([YEAST_IDEA_SOURCE, YEAST_IDEA_TARGET], as_index=True)
        .apply(_summarize_idea_pairs)
        .reset_index()
    )

    # add some more fields are reformat
    formatted_distinct_edges = distinct_edges.rename(
        {YEAST_IDEA_SOURCE: "upstream_name", YEAST_IDEA_TARGET: "downstream_name"},
        axis=1,
    ).assign(
        upstream_compartment="cellular_component",
        downstream_compartment="cellular_component",
        # tag reactions with the IDEA publication
        r_Identifiers=identifiers._format_Identifiers_pubmed(YEAST_IDEA_PUBMED_ID),
        r_isreversible=False,
    )

    # create some nice interaction names before we rename the roles as SBO terms
    formatted_distinct_edges["r_name"] = [
        f"{u} {d} {r} of {t}"
        for u, d, r, t in zip(
            formatted_distinct_edges["upstream_name"],
            formatted_distinct_edges["directness"],
            formatted_distinct_edges["role"],
            formatted_distinct_edges["downstream_name"],
        )
    ]

    # final interaction output
    # replace readable roles with entries in the SBO ontology
    interaction_edgelist = formatted_distinct_edges.replace(
        {"role": MINI_SBO_FROM_NAME}
    ).rename({"role": "sbo_term"}, axis=1)

    species_df = pd.DataFrame(
        {
            "s_name": list(
                {
                    *idea_kinetics_df[YEAST_IDEA_SOURCE],
                    *idea_kinetics_df[YEAST_IDEA_TARGET],
                }
            )
        }
    )

    # create Identifiers objects for each species
    species_df["s_Identifiers"] = [
        identifiers.Identifiers(
            [{"ontology": "gene_name", "identifier": x, "bqb": BQB.IS}]
        )
        for x in species_df["s_name"]
    ]

    # Constant fields (for this data source)

    # setup compartments (just treat this as uncompartmentalized for now)
    compartments_df = sbml_dfs_utils.stub_compartments()

    # Per convention unaggregated models receive an empty source
    interaction_source = source.Source(init=True)

    sbml_dfs = sbml_dfs_core.sbml_dfs_from_edgelist(
        interaction_edgelist=interaction_edgelist,
        species_df=species_df,
        compartments_df=compartments_df,
        interaction_source=interaction_source,
        # additional attributes (directness) are added to reactions_data
        keep_reactions_data="idea",
    )
    sbml_dfs.validate()

    return sbml_dfs


def _summarize_idea_pairs(pairs_data: pd.DataFrame) -> pd.Series:
    """Rollup multiple records of a TF->target pair into a single summary."""

    # specify how to aggregate results if there are more than one entry for a TF-target pair
    # pull most attributes from the earliest change
    # this will favor direct over indirect naturally
    earliest_change = pairs_data.sort_values("t_rise").iloc[0].to_dict()

    KEYS_SUMMARIZED = ["v_inter", "v_final", "t_rise", "t_fall", "rate", "directness"]
    kinetic_timing_dict = {k: earliest_change[k] for k in KEYS_SUMMARIZED}

    # map v_inter (log2 fold-change change following perturbation) onto SBO terms for interactions
    if (any(pairs_data["v_inter"] > 0)) and (any(pairs_data["v_inter"] < 0)):
        kinetic_timing_dict["role"] = "modifier"
    elif all(pairs_data["v_inter"] > 0):
        kinetic_timing_dict["role"] = "stimulator"
    elif all(pairs_data["v_inter"] < 0):
        kinetic_timing_dict["role"] = "inhibitor"
    else:
        ValueError("Unexpected v_inter values")

    return pd.Series(kinetic_timing_dict)



"""
Function execution components for the Napistu MCP server.
"""

from typing import Dict, List, Any, Optional
import inspect
import logging

from fastmcp import FastMCP

from napistu.mcp.component_base import ComponentState, MCPComponent

logger = logging.getLogger(__name__)


class ExecutionState(ComponentState):
    """State management for execution component."""

    def __init__(
        self,
        session_context: Optional[Dict] = None,
        object_registry: Optional[Dict] = None,
    ):
        super().__init__()
        # Session context contains global functions and modules
        self.session_context = session_context or {}
        # Object registry contains user-registered objects
        self.session_objects = object_registry or {}

    def is_healthy(self) -> bool:
        """Component is healthy if it has a session context."""
        return bool(self.session_context)

    def get_health_details(self) -> Dict[str, Any]:
        """Provide execution-specific health details."""
        return {
            "session_context_items": len(self.session_context),
            "registered_objects": len(self.session_objects),
            "context_keys": list(self.session_context.keys()),
            "object_names": list(self.session_objects.keys()),
        }

    def register_object(self, name: str, obj: Any) -> None:
        """Register an object with the execution component."""
        self.session_objects[name] = obj
        logger.info(f"Registered object '{name}' with MCP server")


class ExecutionComponent(MCPComponent):
    """MCP component for function execution and object management."""

    def __init__(
        self,
        session_context: Optional[Dict] = None,
        object_registry: Optional[Dict] = None,
    ):
        # Override parent constructor to pass context to state
        self.state = ExecutionState(session_context, object_registry)

    def _create_state(self) -> ExecutionState:
        """This won't be called due to overridden constructor."""
        pass

    async def initialize(self) -> bool:
        """
        Initialize execution component by setting up the session context.

        Returns
        -------
        bool
            True if initialization successful
        """
        try:
            # Import and add napistu to session context
            import napistu

            self.state.session_context["napistu"] = napistu

            logger.info("Execution component initialized with napistu module")
            return True

        except ImportError as e:
            logger.error(f"Failed to import napistu module: {e}")
            return False

    def register_object(self, name: str, obj: Any) -> None:
        """
        Register an object with the execution component.

        Args:
            name: Name to reference the object by
            obj: The object to register
        """
        self.state.register_object(name, obj)

    def register(self, mcp: FastMCP) -> None:
        """
        Register execution resources and tools with the MCP server.

        Parameters
        ----------
        mcp : FastMCP
            FastMCP server instance
        """

        # Register resources
        @mcp.resource("napistu://execution/registry")
        async def get_registry():
            """Get a summary of all objects registered with the server."""
            return {
                "object_count": len(self.state.session_objects),
                "object_names": list(self.state.session_objects.keys()),
                "object_types": {
                    name: type(obj).__name__
                    for name, obj in self.state.session_objects.items()
                },
            }

        @mcp.resource("napistu://execution/environment")
        async def get_environment_info() -> Dict[str, Any]:
            """Get information about the local Python environment."""
            try:
                import napistu

                napistu_version = getattr(napistu, "__version__", "unknown")
            except ImportError:
                napistu_version = "not installed"

            import sys

            return {
                "python_version": sys.version,
                "napistu_version": napistu_version,
                "platform": sys.platform,
                "registered_objects": list(self.state.session_objects.keys()),
                "session_context": list(self.state.session_context.keys()),
            }

        # Register tools
        @mcp.tool()
        async def list_registry() -> Dict[str, Any]:
            """List all objects registered with the server."""
            result = {}

            for name, obj in self.state.session_objects.items():
                obj_type = type(obj).__name__

                # Get additional info based on object type
                if hasattr(obj, "shape"):  # For pandas DataFrame or numpy array
                    obj_info = {
                        "type": obj_type,
                        "shape": str(obj.shape),
                    }
                elif hasattr(obj, "__len__"):  # For lists, dicts, etc.
                    obj_info = {
                        "type": obj_type,
                        "length": len(obj),
                    }
                else:
                    obj_info = {
                        "type": obj_type,
                    }

                result[name] = obj_info

            return result

        @mcp.tool()
        async def describe_object(object_name: str) -> Dict[str, Any]:
            """Get detailed information about a registered object."""
            if object_name not in self.state.session_objects:
                return {"error": f"Object '{object_name}' not found in registry"}

            obj = self.state.session_objects[object_name]
            obj_type = type(obj).__name__

            # Basic info for all objects
            result = {
                "name": object_name,
                "type": obj_type,
                "methods": [],
                "attributes": [],
            }

            # Add methods and attributes
            for name in dir(obj):
                if name.startswith("_"):
                    continue

                try:
                    attr = getattr(obj, name)

                    if callable(attr):
                        # Method
                        sig = str(inspect.signature(attr))
                        doc = inspect.getdoc(attr) or ""
                        result["methods"].append(
                            {
                                "name": name,
                                "signature": sig,
                                "docstring": doc,
                            }
                        )
                    else:
                        # Attribute
                        attr_type = type(attr).__name__
                        result["attributes"].append(
                            {
                                "name": name,
                                "type": attr_type,
                            }
                        )
                except Exception:
                    # Skip attributes that can't be accessed
                    pass

            return result

        @mcp.tool()
        async def execute_function(
            function_name: str,
            object_name: Optional[str] = None,
            args: Optional[List] = None,
            kwargs: Optional[Dict] = None,
        ) -> Dict[str, Any]:
            """Execute a Napistu function on a registered object."""
            args = args or []
            kwargs = kwargs or {}

            try:
                if object_name:
                    # Method call on an object
                    if object_name not in self.state.session_objects:
                        return {
                            "error": f"Object '{object_name}' not found in registry"
                        }

                    obj = self.state.session_objects[object_name]

                    if not hasattr(obj, function_name):
                        return {
                            "error": f"Method '{function_name}' not found on object '{object_name}'"
                        }

                    func = getattr(obj, function_name)
                    result = func(*args, **kwargs)
                else:
                    # Global function call
                    if function_name in self.state.session_context:
                        # Function from session context
                        func = self.state.session_context[function_name]
                        result = func(*args, **kwargs)
                    else:
                        # Try to find the function in Napistu
                        try:
                            import napistu

                            # Split function name by dots for nested modules
                            parts = function_name.split(".")
                            current = napistu

                            for part in parts[:-1]:
                                current = getattr(current, part)

                            func = getattr(current, parts[-1])
                            result = func(*args, **kwargs)
                        except (ImportError, AttributeError):
                            return {"error": f"Function '{function_name}' not found"}

                # Register result if it's a return value
                if result is not None:
                    result_name = f"result_{len(self.state.session_objects) + 1}"
                    self.state.session_objects[result_name] = result

                    # Basic type conversion for JSON serialization
                    if hasattr(result, "to_dict"):
                        # For pandas DataFrame or similar
                        return {
                            "success": True,
                            "result_name": result_name,
                            "result_type": type(result).__name__,
                            "result_preview": (
                                result.to_dict()
                                if hasattr(result, "__len__") and len(result) < 10
                                else "Result too large to preview"
                            ),
                        }
                    elif hasattr(result, "to_json"):
                        # For objects with JSON serialization
                        return {
                            "success": True,
                            "result_name": result_name,
                            "result_type": type(result).__name__,
                            "result_preview": result.to_json(),
                        }
                    elif hasattr(result, "__dict__"):
                        # For custom objects
                        return {
                            "success": True,
                            "result_name": result_name,
                            "result_type": type(result).__name__,
                            "result_preview": str(result),
                        }
                    else:
                        # For simple types
                        return {
                            "success": True,
                            "result_name": result_name,
                            "result_type": type(result).__name__,
                            "result_preview": str(result),
                        }
                else:
                    return {
                        "success": True,
                        "result": None,
                    }
            except Exception as e:
                import traceback

                return {
                    "error": str(e),
                    "traceback": traceback.format_exc(),
                }

        @mcp.tool()
        async def search_paths(
            source_node: str,
            target_node: str,
            network_object: str,
            max_depth: int = 3,
        ) -> Dict[str, Any]:
            """Find paths between two nodes in a network."""
            if network_object not in self.state.session_objects:
                return {
                    "error": f"Network object '{network_object}' not found in registry"
                }

            network = self.state.session_objects[network_object]

            try:
                # Import necessary modules
                import napistu

                # Check if the object is a valid network type
                if hasattr(network, "find_paths"):
                    # Direct method call
                    paths = network.find_paths(
                        source_node, target_node, max_depth=max_depth
                    )
                elif hasattr(napistu.graph, "find_paths"):
                    # Function call
                    paths = napistu.graph.find_paths(
                        network, source_node, target_node, max_depth=max_depth
                    )
                else:
                    return {"error": "Could not find appropriate path-finding function"}

                # Register result
                result_name = f"paths_{len(self.state.session_objects) + 1}"
                self.state.session_objects[result_name] = paths

                # Return results
                if hasattr(paths, "to_dict"):
                    return {
                        "success": True,
                        "result_name": result_name,
                        "paths_found": (
                            len(paths) if hasattr(paths, "__len__") else "unknown"
                        ),
                        "result_preview": (
                            paths.to_dict()
                            if hasattr(paths, "__len__") and len(paths) < 10
                            else "Result too large to preview"
                        ),
                    }
                else:
                    return {
                        "success": True,
                        "result_name": result_name,
                        "paths_found": (
                            len(paths) if hasattr(paths, "__len__") else "unknown"
                        ),
                        "result_preview": str(paths),
                    }
            except Exception as e:
                import traceback

                return {
                    "error": str(e),
                    "traceback": traceback.format_exc(),
                }


# Module-level component instance (will be created by server with proper context)
_component: Optional[ExecutionComponent] = None


def create_component(
    session_context: Optional[Dict] = None, object_registry: Optional[Dict] = None
) -> ExecutionComponent:
    """
    Create and configure the execution component with session context.

    Args:
        session_context: Dictionary of the user's current session (e.g., globals())
        object_registry: Dictionary of named objects to make available

    Returns:
        ExecutionComponent: Configured execution component
    """
    global _component
    _component = ExecutionComponent(session_context, object_registry)
    return _component


def get_component() -> ExecutionComponent:
    """
    Get the execution component instance.

    Returns
    -------
    ExecutionComponent
        The execution component instance

    Raises
    ------
    RuntimeError
        If component hasn't been created yet
    """
    if _component is None:
        raise RuntimeError(
            "Execution component not created. Call create_component() first."
        )
    return _component


def register_object(name: str, obj: Any) -> None:
    """
    Register an object with the execution component (legacy function).

    Args:
        name: Name to reference the object by
        obj: The object to register
    """
    if _component is None:
        raise RuntimeError(
            "Execution component not created. Call create_component() first."
        )
    _component.register_object(name, obj)



from __future__ import annotations

import logging
import math

import numpy as np
import pandas as pd

from napistu.network.ng_core import NapistuGraph
from napistu.network.ig_utils import validate_edge_attributes
from napistu.constants import NAPISTU_EDGELIST, SBML_DFS
from napistu.network.constants import (
    NAPISTU_GRAPH_EDGES,
)

logger = logging.getLogger(__name__)


def precompute_distances(
    napistu_graph: NapistuGraph,
    max_steps: int = -1,
    max_score_q: float = float(1),
    partition_size: int = int(5000),
    weights_vars: list[str] = [
        NAPISTU_GRAPH_EDGES.WEIGHTS,
        NAPISTU_GRAPH_EDGES.UPSTREAM_WEIGHTS,
    ],
) -> pd.DataFrame:
    """
    Precompute Distances between all pairs of species in a NapistuGraph network.

    Parameters
    ----------
    napistu_graph: NapistuGraph
        An NapistuGraph network model (subclass of igraph.Graph)
    max_steps: int
        The maximum number of steps between pairs of species to save a distance
    max_score_q: float
        Retain up to the "max_score_q" quantiles of all scores (small scores are better)
    partition_size: int
        The number of species to process together when computing distances. Decreasing this
        value will lower the overall memory footprint of distance calculation.
    weights_vars: list
        One or more variables defining edge weights to use when calculating weighted
        shortest paths. Shortest paths will be separately calculated with each type of
        weights and used to construct path weights named according to 'path_{weight_var}'

    Returns:
    ----------
    A pd.DataFrame containing:
    - sc_id_origin: origin node
    - sc_id_dest: destination node
    - path_length: minimum path length between from and to
    - path_weight*: minimum path weight between from and to (formed by summing the weights of individual edges).
      *One variable will exist for each weight specified in 'weights_vars'

    """

    if max_steps == -1:
        max_steps = int(100000)

    # validate inputs
    if max_steps < 1:
        raise ValueError(f"max_steps must >= 1, but was {max_steps}")

    if (max_score_q < 0) or (max_score_q > 1):
        raise ValueError(f"max_score_q must be between 0 and 1 but was {max_score_q}")

    # make sure weight vars exist
    validate_edge_attributes(napistu_graph, weights_vars)

    # assign molecular species to partitions
    vs_to_partition = pd.DataFrame(
        {"sc_id": napistu_graph.vs["name"], "node_type": napistu_graph.vs["node_type"]}
    ).query("node_type == 'species'")

    n_paritions = math.ceil(vs_to_partition.shape[0] / partition_size)

    vs_to_partition["partition"] = vs_to_partition.index % n_paritions
    vs_to_partition = vs_to_partition.set_index("partition").sort_index()

    # interate through all partitions of "from" nodes and find their shortest and lowest weighted paths
    unique_partitions = vs_to_partition.index.unique().tolist()

    logger.info(f"Calculating distances for {len(unique_partitions)} partitions")
    precomputed_distances = pd.concat(
        [
            _calculate_distances_subset(
                napistu_graph,
                vs_to_partition,
                vs_to_partition.loc[uq_part],
                weights_vars=weights_vars,
            )
            for uq_part in unique_partitions
        ]
    ).query("sc_id_origin != sc_id_dest")

    # filter by path length and/or weight

    logger.info(
        f"Filtering distances by path length ({max_steps}) and score quantile ({max_score_q})"
    )
    filtered_precomputed_distances = _filter_precomputed_distances(
        precomputed_distances=precomputed_distances,
        max_steps=max_steps,
        max_score_q=max_score_q,
        path_weights_vars=["path_" + w for w in weights_vars],
    ).reset_index(drop=True)

    return filtered_precomputed_distances


def _calculate_distances_subset(
    napistu_graph: NapistuGraph,
    vs_to_partition: pd.DataFrame,
    one_partition: pd.DataFrame,
    weights_vars: list[str] = [
        NAPISTU_GRAPH_EDGES.WEIGHTS,
        NAPISTU_GRAPH_EDGES.UPSTREAM_WEIGHTS,
    ],
) -> pd.DataFrame:
    """Calculate distances from a subset of vertices to all vertices."""

    d_steps = (
        pd.DataFrame(
            np.array(
                napistu_graph.distances(
                    source=one_partition[SBML_DFS.SC_ID],
                    target=vs_to_partition[SBML_DFS.SC_ID],
                )
            ),
            index=one_partition[SBML_DFS.SC_ID].rename(NAPISTU_EDGELIST.SC_ID_ORIGIN),
            columns=vs_to_partition[SBML_DFS.SC_ID].rename(NAPISTU_EDGELIST.SC_ID_DEST),
        )
        .reset_index()
        .melt(NAPISTU_EDGELIST.SC_ID_ORIGIN, value_name="path_length")
        .replace([np.inf, -np.inf], np.nan, inplace=False)
        .dropna()
    )

    d_weights_list = list()
    for weight_type in weights_vars:
        d_weights_subset = (
            pd.DataFrame(
                np.array(
                    napistu_graph.distances(
                        source=one_partition[SBML_DFS.SC_ID],
                        target=vs_to_partition[SBML_DFS.SC_ID],
                        weights=weight_type,
                    )
                ),
                index=one_partition[SBML_DFS.SC_ID].rename(
                    NAPISTU_EDGELIST.SC_ID_ORIGIN
                ),
                columns=vs_to_partition[SBML_DFS.SC_ID].rename(
                    NAPISTU_EDGELIST.SC_ID_DEST
                ),
            )
            .reset_index()
            .melt(NAPISTU_EDGELIST.SC_ID_ORIGIN, value_name=f"path_{weight_type}")
            .replace([np.inf, -np.inf], np.nan, inplace=False)
            .dropna()
        )

        d_weights_list.append(d_weights_subset)

    d_weights = d_weights_list.pop()
    while len(d_weights_list) > 0:
        d_weights = d_weights.merge(d_weights_list.pop())

    # merge shortest path distances by length and by weight
    # note: these may be different paths! e.g., a longer path may have a lower weight than a short one
    path_summaries = d_steps.merge(
        d_weights,
        left_on=[NAPISTU_EDGELIST.SC_ID_ORIGIN, NAPISTU_EDGELIST.SC_ID_DEST],
        right_on=[NAPISTU_EDGELIST.SC_ID_ORIGIN, NAPISTU_EDGELIST.SC_ID_DEST],
    )

    # return connected species
    return path_summaries


def _filter_precomputed_distances(
    precomputed_distances: pd.DataFrame,
    max_steps: float | int = np.inf,
    max_score_q: float = 1,
    path_weights_vars: list[str] = ["path_weights", "path_upstream_weights"],
) -> pd.DataFrame:
    """Filter precomputed distances by maximum steps and/or to low scores by quantile."""

    # filter by path lengths
    short_precomputed_distances = precomputed_distances[
        precomputed_distances["path_length"] <= max_steps
    ]
    n_filtered_by_path_length = (
        precomputed_distances.shape[0] - short_precomputed_distances.shape[0]
    )
    if n_filtered_by_path_length > 0:
        logger.info(
            f"filtered {n_filtered_by_path_length} possible paths with length > {max_steps}"
        )

    # filter by path weights
    for wt_var in path_weights_vars:
        score_q_cutoff = np.quantile(short_precomputed_distances[wt_var], max_score_q)

        short_precomputed_distances.loc[
            short_precomputed_distances[wt_var] > score_q_cutoff, wt_var
        ] = np.nan

    valid_weights = short_precomputed_distances[path_weights_vars].dropna(how="all")

    low_weight_precomputed_distances = short_precomputed_distances[
        short_precomputed_distances.index.isin(valid_weights.index.tolist())
    ]

    n_filtered_by_low_weight = (
        short_precomputed_distances.shape[0] - low_weight_precomputed_distances.shape[0]
    )

    if n_filtered_by_low_weight > 0:
        logger.info(
            f"filtered {n_filtered_by_low_weight} possible paths with path weights greater"
        )
        logger.info(f"than the {max_score_q} quantile of the path weight distribution")

    weight_nan_summary = valid_weights.isnull().sum()
    if any(weight_nan_summary != 0):
        nan_summary = " and ".join(
            [
                f"{k} has {v} np.nan values"
                for k, v in weight_nan_summary.to_dict().items()
            ]
        )
        logger.info(nan_summary)

    return low_weight_precomputed_distances



from __future__ import annotations

import logging
import os
import re
from typing import Any

import libsbml
import pandas as pd
from fs import open_fs
from pydantic import field_validator, RootModel

from napistu import consensus
from napistu import identifiers
from napistu import sbml_dfs_utils
from napistu import source
from napistu import utils
from napistu.constants import BQB
from napistu.constants import ONTOLOGIES
from napistu.constants import SBML_DFS
from napistu.constants import SBML_DFS_SCHEMA
from napistu.constants import SCHEMA_DEFS
from napistu.ingestion.constants import SBML_DEFS
from napistu.ingestion.constants import COMPARTMENTS_GO_TERMS
from napistu.ingestion.constants import COMPARTMENT_ALIASES
from napistu.ingestion.constants import VALID_COMPARTMENTS
from napistu.ingestion.constants import GENERIC_COMPARTMENT

logger = logging.getLogger(__name__)


class SBML:
    """A class for handling Systems Biology Markup Language (SBML) files.

    This class provides an interface to read and parse SBML files, offering
    methods to access the model, summarize its contents, and report any errors
    encountered during parsing.

    Parameters
    ----------
    sbml_path : str
        The file path to an SBML model. Supports local paths and GCS URIs.

    Attributes
    ----------
    document : libsbml.SBMLDocument
        The raw SBML document object from libsbml.
    model : libsbml.Model
        The parsed SBML model object from libsbml.

    Methods
    -------
    summary()
        Prints a summary of the sbml model
    sbml_errors(reduced_log, return_df)
        Print a summary of all errors in the SBML file

    Raises
    ------
    ValueError
        If the SBML model is not Level 3, or if critical, unknown errors are
        found during parsing.
    """

    def __init__(
        self,
        sbml_path: str,
    ) -> None:
        """Initializes the SBML object by reading and validating an SBML file."""
        reader = libsbml.SBMLReader()
        if os.path.exists(sbml_path):
            self.document = reader.readSBML(sbml_path)
        else:
            with open_fs(os.path.dirname(sbml_path)) as fs:
                txt = fs.readtext(os.path.basename(sbml_path))
                self.document = reader.readSBMLFromString(txt)

        if self.document.getLevel() < 3:
            raise ValueError(
                f"SBML model is level {self.document.getLevel()}, only SBML 3 is supported"
            )

        self.model = self.document.getModel()

        # check for critical sbml errors
        errors = self.sbml_errors(reduced_log=False, return_df=True)
        if errors is not None:
            critical_errors = errors[errors[SBML_DEFS.ERROR_SEVERITY] >= 2]
            critical_errors = set(critical_errors[SBML_DEFS.ERROR_DESCRIPTION].unique())
            known_errors = {"<layout> must have 'id' and may have 'name'"}

            found_known_errors = known_errors.intersection(critical_errors)
            if len(found_known_errors) > 0:
                logger.warning(
                    f"The following known errors were found: {found_known_errors}"
                )

            unknown_critical_errors = critical_errors - known_errors
            if len(unknown_critical_errors) != 0:
                critical_errors = ", ".join(unknown_critical_errors)
                raise ValueError(
                    f"Critical errors were found when reading the sbml file: {critical_errors}"
                )

    def sbml_errors(self, reduced_log: bool = True, return_df: bool = False):
        """Formats and reports all errors found in the SBML file.

        Parameters
        ----------
        reduced_log : bool, optional
            If True, aggregates errors by category and severity. Defaults to True.
        return_df : bool, optional
            If True, returns a DataFrame of the errors. Otherwise, prints a
            styled summary. Defaults to False.

        Returns
        -------
        pd.DataFrame or None
            A DataFrame containing the error log if `return_df` is True and
            errors are present, otherwise None.
        """
        n_errors = self.document.getNumErrors()
        if n_errors == 0:
            return None

        error_log = list()
        for i in range(n_errors):
            e = self.document.getError(i)

            error_entry = {
                SBML_DEFS.ERROR_NUMBER: i,
                SBML_DEFS.ERROR_CATEGORY: e.getCategoryAsString(),
                SBML_DEFS.ERROR_SEVERITY: e.getSeverity(),
                SBML_DEFS.ERROR_DESCRIPTION: e.getShortMessage(),
                SBML_DEFS.ERROR_MESSAGE: e.getMessage(),
            }

            error_log.append(error_entry)
        error_log = pd.DataFrame(error_log)

        if reduced_log:
            error_log = (
                error_log[
                    [
                        SBML_DEFS.ERROR_CATEGORY,
                        SBML_DEFS.ERROR_SEVERITY,
                        SBML_DEFS.ERROR_MESSAGE,
                    ]
                ]
                .groupby([SBML_DEFS.ERROR_CATEGORY, SBML_DEFS.ERROR_SEVERITY])
                .count()
            )

        if return_df:
            return error_log
        else:
            if reduced_log:
                headers = [
                    f"{SBML_DEFS.ERROR_CATEGORY}, {SBML_DEFS.ERROR_SEVERITY}",
                    "count",
                ]
            else:
                headers = [
                    SBML_DEFS.ERROR_CATEGORY,
                    SBML_DEFS.ERROR_SEVERITY,
                    SBML_DEFS.ERROR_DESCRIPTION,
                ]
                error_log = error_log[headers]

            utils.style_df(error_log, headers=headers)

            return None

    def summary(self) -> pd.DataFrame:
        """Generates a styled summary of the SBML model.

        Returns
        -------
        pd.io.formats.style.Styler
            A styled pandas DataFrame containing a summary of the model,
            including pathway name, ID, and counts of species and reactions.
        """
        model = self.model

        model_summaries = dict()

        model_summaries[SBML_DEFS.SUMMARY_PATHWAY_NAME] = model.getName()
        model_summaries[SBML_DEFS.SUMMARY_PATHWAY_ID] = model.getId()

        model_summaries[SBML_DEFS.SUMMARY_N_SPECIES] = model.getNumSpecies()
        model_summaries[SBML_DEFS.SUMMARY_N_REACTIONS] = model.getNumReactions()

        compartments = [
            model.getCompartment(i).getName() for i in range(model.getNumCompartments())
        ]
        compartments.sort()
        model_summaries[SBML_DEFS.SUMMARY_COMPARTMENTS] = ",\n".join(compartments)

        model_summaries_dat = pd.DataFrame(model_summaries, index=[0]).T

        return utils.style_df(model_summaries_dat)  # type: ignore

    def _define_compartments(
        self, compartment_aliases_dict: dict | None = None
    ) -> pd.DataFrame:
        """Extracts and defines compartments from the SBML model.

        This function iterates through the compartments in the SBML model,
        extracting their IDs, names, and identifiers. It also handles cases where
        CVTerms are missing by mapping compartment names to known GO terms.

        Parameters
        ----------
        sbml_model : SBML
            The SBML model to process.
        compartment_aliases_dict : dict, optional
            A dictionary to map custom compartment names. If None, the default
            mapping from `COMPARTMENT_ALIASES` is used.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing information about each compartment, indexed by
            compartment ID.
        """
        if compartment_aliases_dict is None:
            aliases = COMPARTMENT_ALIASES
        else:
            aliases = CompartmentAliasesValidator.from_dict(compartment_aliases_dict)

        compartments = list()
        for i in range(self.model.getNumCompartments()):
            comp = self.model.getCompartment(i)

            if not comp.getCVTerms():
                logger.warning(
                    f"Compartment {comp.getId()} has empty CVterms, mapping its c_Identifiers from the Compartment dict"
                )

                compartments.append(_define_compartments_missing_cvterms(comp, aliases))

            else:
                compartments.append(
                    {
                        SBML_DFS.C_ID: comp.getId(),
                        SBML_DFS.C_NAME: comp.getName(),
                        SBML_DFS.C_IDENTIFIERS: identifiers.cv_to_Identifiers(comp),
                        SBML_DFS.C_SOURCE: source.Source(init=True),
                    }
                )

        return pd.DataFrame(compartments).set_index(SBML_DFS.C_ID)

    def _define_cspecies(self) -> pd.DataFrame:
        """Creates a DataFrame of compartmentalized species from an SBML model.

        This function extracts all species from the model and creates a
        standardized DataFrame that includes unique IDs for each compartmentalized
        species (`sc_id`), along with species and compartment IDs, and their
        corresponding identifiers.

        Returns
        -------
        pd.DataFrame
            A DataFrame containing information about each compartmentalized species.
        """
        comp_species = list()
        for i in range(self.model.getNumSpecies()):
            spec = self.model.getSpecies(i)

            spec_dict = {
                SBML_DFS.SC_ID: spec.getId(),
                SBML_DFS.SC_NAME: spec.getName(),
                SBML_DFS.C_ID: spec.getCompartment(),
                SBML_DFS.S_IDENTIFIERS: identifiers.cv_to_Identifiers(spec),
                SBML_DFS.SC_SOURCE: source.Source(init=True),
            }

            comp_species.append(spec_dict)

        # add geneproducts defined using L3 FBC extension
        fbc_gene_products = self._define_fbc_gene_products()
        comp_species.extend(fbc_gene_products)

        comp_species_df = pd.DataFrame(comp_species).set_index(SBML_DFS.SC_ID)
        comp_species_df[SBML_DFS.SC_NAME] = utils.update_pathological_names(
            comp_species_df[SBML_DFS.SC_NAME], "SC"
        )

        return comp_species_df

    def _define_fbc_gene_products(self) -> list[dict]:

        mplugin = self.model.getPlugin("fbc")

        fbc_gene_products = list()
        if mplugin is not None:
            for i in range(mplugin.getNumGeneProducts()):
                gene_product = mplugin.getGeneProduct(i)

                gene_dict = {
                    SBML_DFS.SC_ID: gene_product.getId(),
                    SBML_DFS.SC_NAME: (
                        gene_product.getName()
                        if gene_product.isSetName()
                        else gene_product.getLabel()
                    ),
                    # use getLabel() to accomendate sbml model (e.g. HumanGEM.xml) with no fbc:name attribute
                    # Recon3D.xml has both fbc:label and fbc:name attributes, with gene name in fbc:nam
                    SBML_DFS.C_ID: None,
                    SBML_DFS.S_IDENTIFIERS: identifiers.cv_to_Identifiers(gene_product),
                    SBML_DFS.SC_SOURCE: source.Source(init=True),
                }

                fbc_gene_products.append(gene_dict)

        return fbc_gene_products

    def _define_reactions(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Extracts and defines reactions and their participating species.

        This function iterates through all reactions in the SBML model, creating
        a DataFrame for reaction attributes and another for all participating
        species (reactants, products, and modifiers).

        Parameters
        ----------
        sbml_model : SBML
            The SBML model to process.

        Returns
        -------
        tuple[pd.DataFrame, pd.DataFrame]
            A tuple containing two DataFrames:
            - The first DataFrame contains reaction attributes, indexed by reaction ID.
            - The second DataFrame lists all species participating in reactions.
        """
        reactions_list = []
        reaction_species_list = []
        for i in range(self.model.getNumReactions()):
            rxn = SBML_reaction(self.model.getReaction(i))
            reactions_list.append(rxn.reaction_dict)

            rxn_specs = rxn.species
            rxn_specs[SBML_DFS.R_ID] = rxn.reaction_dict[SBML_DFS.R_ID]
            reaction_species_list.append(rxn_specs)

        reactions = pd.DataFrame(reactions_list).set_index(SBML_DFS.R_ID)

        reaction_species_df = pd.concat(reaction_species_list)
        # add an index if reaction species didn't have IDs in the .sbml
        if all([v == "" for v in reaction_species_df.index.tolist()]):
            reaction_species_df = (
                reaction_species_df.reset_index(drop=True)
                .assign(
                    rsc_id=sbml_dfs_utils.id_formatter(
                        range(reaction_species_df.shape[0]), SBML_DFS.RSC_ID
                    )
                )
                .set_index(SBML_DFS.RSC_ID)
            )

        return reactions, reaction_species_df

    def _define_species(self) -> tuple[pd.DataFrame, pd.DataFrame]:
        """Extracts and defines species and compartmentalized species.

        This function creates two DataFrames: one for unique molecular species
        (un-compartmentalized) and another for compartmentalized species, which
        represent a species within a specific compartment.

        Returns
        -------
        tuple[pd.DataFrame, pd.DataFrame]
            A tuple containing two DataFrames:
            - The first DataFrame represents unique molecular species.
            - The second DataFrame represents compartmentalized species.
        """

        SPECIES_SCHEMA = SBML_DFS_SCHEMA.SCHEMA[SBML_DFS.SPECIES]
        CSPECIES_SCHEMA = SBML_DFS_SCHEMA.SCHEMA[SBML_DFS.COMPARTMENTALIZED_SPECIES]
        SPECIES_VARS = SPECIES_SCHEMA[SCHEMA_DEFS.VARS]
        CSPECIES_VARS = CSPECIES_SCHEMA[SCHEMA_DEFS.VARS]

        comp_species_df = self._define_cspecies()

        # find unique species and create a table
        consensus_species_df = comp_species_df.copy()
        consensus_species_df.index.names = [SBML_DFS.S_ID]
        consensus_species, species_lookup = consensus.reduce_to_consensus_ids(
            consensus_species_df,
            # note that this is an incomplete schema because consensus_species_df isn't a
            # normal species table
            {
                SCHEMA_DEFS.PK: SBML_DFS.S_ID,
                SCHEMA_DEFS.ID: SBML_DFS.S_IDENTIFIERS,
                SCHEMA_DEFS.TABLE: SBML_DFS.SPECIES,
            },
        )

        # create a table of unique molecular species
        consensus_species.index.name = SBML_DFS.S_ID
        consensus_species[SBML_DFS.S_NAME] = [
            re.sub("\\[.+\\]", "", x).strip()
            for x in consensus_species[SBML_DFS.SC_NAME]
        ]
        consensus_species = consensus_species.drop(
            [SBML_DFS.SC_NAME, SBML_DFS.C_ID], axis=1
        )
        consensus_species[SBML_DFS.S_SOURCE] = [
            source.Source(init=True) for x in range(0, consensus_species.shape[0])
        ]

        species = consensus_species[SPECIES_VARS]
        compartmentalized_species = comp_species_df.join(species_lookup).rename(
            columns={"new_id": SBML_DFS.S_ID}
        )[CSPECIES_VARS]

        return species, compartmentalized_species


class CompartmentAliasesValidator(RootModel):
    """
    A Pydantic model for validating compartment alias dictionaries.

    This model ensures that the compartment alias dictionary is a mapping
    from a string (the canonical compartment name) to a list of strings
    (the aliases for that compartment). It also validates that the keys
    of the dictionary are valid compartment names.

    Attributes
    ----------
    root : dict[str, list[str]]
        The root of the model is a dictionary where keys are strings and
        values are lists of strings.
    """

    root: dict[str, list[str]]

    @field_validator("root")
    def validate_aliases(cls, values: dict[str, list[str]]):
        """Validate the compartment alias dictionary."""
        for key, alias_list in values.items():
            if not key:
                raise ValueError("Compartment keys must be non-empty.")
            if key not in VALID_COMPARTMENTS:
                raise ValueError(
                    f"Invalid compartment key: {key}. "
                    f"Must be one of {VALID_COMPARTMENTS}"
                )
            if not alias_list:
                raise ValueError(f"Alias list for '{key}' cannot be empty.")
        return values

    @classmethod
    def from_dict(cls, data: dict[str, list[str]]) -> "CompartmentAliasesValidator":
        """
        Create a CompartmentAliasesValidator from a dictionary.

        Parameters
        ----------
        data : dict[str, list[str]]
            A dictionary mapping canonical compartment names to their aliases.

        Returns
        -------
        CompartmentAliasesValidator
            A validated instance of the model.
        """
        return cls.model_validate(data)

    def __getitem__(self, key: str) -> list[str]:
        return self.root[key]

    def items(self):
        return self.root.items()

    def __iter__(self):
        return iter(self.root)

    def __len__(self):
        return len(self.root)


class SBML_reaction:
    """A convenience class for processing individual SBML reactions.

    This class extracts and organizes key information about an SBML reaction,
    including its attributes and participating species (substrates, products,
    and modifiers).

    Parameters
    ----------
    sbml_reaction : libsbml.Reaction
        A libsbml Reaction object to be processed.

    Attributes
    ----------
    reaction_dict : dict
        A dictionary of reaction-level attributes, including its ID, name,
        reversibility, identifiers, and source information.
    species : pd.DataFrame
        A DataFrame listing all species participating in the reaction,
        including their roles (substrate, product, modifier), stoichiometry,
        and SBO terms.
    """

    def __init__(
        self,
        sbml_reaction: libsbml.Reaction,
    ) -> None:
        """Initializes the SBML_reaction object by parsing a libsbml Reaction."""
        reaction_dict = {
            SBML_DFS.R_ID: sbml_reaction.getId(),
            SBML_DFS.R_NAME: sbml_reaction.getName(),
            SBML_DFS.R_IDENTIFIERS: identifiers.cv_to_Identifiers(sbml_reaction),
            SBML_DFS.R_SOURCE: source.Source(init=True),
            SBML_DFS.R_ISREVERSIBLE: sbml_reaction.getReversible(),
        }

        self.reaction_dict = reaction_dict

        # process reaction species
        reaction_species = list()
        # save modifiers
        for i in range(sbml_reaction.getNumModifiers()):
            spec = sbml_reaction.getModifier(i)
            spec_dict = {
                SBML_DFS.RSC_ID: spec.getId(),
                SBML_DFS.SC_ID: spec.getSpecies(),
                SBML_DFS.STOICHIOMETRY: 0,
                SBML_DFS.SBO_TERM: spec.getSBOTermID(),
            }
            reaction_species.append(spec_dict)

        # find gene products defined using the fbc plugin
        rxn_fbc = sbml_reaction.getPlugin("fbc")
        if rxn_fbc:
            gpa = rxn_fbc.getGeneProductAssociation()
            if gpa:
                gene_products = _extract_gene_products(gpa.getAssociation())
                # de-duplicate
                gene_products = list(
                    {d[SBML_DFS.SC_ID]: d for d in gene_products}.values()
                )
                reaction_species.extend(gene_products)

        # save reactants
        for i in range(sbml_reaction.getNumReactants()):
            spec = sbml_reaction.getReactant(i)
            spec_dict = {
                SBML_DFS.RSC_ID: spec.getId(),
                SBML_DFS.SC_ID: spec.getSpecies(),
                SBML_DFS.STOICHIOMETRY: -1 * spec.getStoichiometry(),
                SBML_DFS.SBO_TERM: spec.getSBOTermID(),
            }
            reaction_species.append(spec_dict)
        # save products
        for i in range(sbml_reaction.getNumProducts()):
            spec = sbml_reaction.getProduct(i)
            spec_dict = {
                SBML_DFS.RSC_ID: spec.getId(),
                SBML_DFS.SC_ID: spec.getSpecies(),
                SBML_DFS.STOICHIOMETRY: spec.getStoichiometry(),
                SBML_DFS.SBO_TERM: spec.getSBOTermID(),
            }
            reaction_species.append(spec_dict)

        self.species = pd.DataFrame(reaction_species).set_index(SBML_DFS.RSC_ID)


def sbml_dfs_from_sbml(self, sbml_model: SBML, compartment_aliases: dict | None = None):
    """Parses an SBML model into a set of standardized DataFrames.

    This function serves as the main entry point for converting an SBML model
    into the internal DataFrame-based representation used by napistu. It
    orchestrates the processing of compartments, species, and reactions.

    Parameters
    ----------
    self : object
        The instance of the calling class, expected to have a `schema` attribute.
    sbml_model : SBML
        The SBML model to be parsed.
    compartment_aliases : dict, optional
        A dictionary to map custom compartment names to the napistu controlled
        vocabulary. If None, the default mapping (COMPARTMENT_ALIASES) is used.
        Defaults to None.

    Returns
    -------
    object
        The calling class instance, now populated with DataFrames for
        compartments, species, compartmentalized_species, reactions, and reaction_species
    """
    # 1. Process compartments from the SBML model
    self.compartments = sbml_model._define_compartments(compartment_aliases)

    # 2. Process species and compartmentalized species
    self.species, self.compartmentalized_species = sbml_model._define_species()

    # 3. Process reactions and their participating species
    self.reactions, self.reaction_species = sbml_model._define_reactions()

    return self


def _define_compartments_missing_cvterms(
    comp: libsbml.Compartment, aliases: dict
) -> dict[str, Any]:

    comp_name = comp.getName()
    mapped_compartment_key = [
        compkey for compkey, mappednames in aliases.items() if comp_name in mappednames
    ]

    if len(mapped_compartment_key) == 0:
        logger.warning(
            f"No GO compartment for {comp_name} is mapped, use the generic cellular_component's GO id"
        )

        compartment_entry = {
            SBML_DFS.C_ID: comp.getId(),
            SBML_DFS.C_NAME: comp.getName(),
            SBML_DFS.C_IDENTIFIERS: identifiers.Identifiers(
                [
                    identifiers.format_uri(
                        uri=identifiers.create_uri_url(
                            ontology=ONTOLOGIES.GO,
                            identifier=COMPARTMENTS_GO_TERMS[GENERIC_COMPARTMENT],
                        ),
                        biological_qualifier_type=BQB.BQB_IS,
                    )
                ]
            ),
            SBML_DFS.C_SOURCE: source.Source(init=True),
        }

    if len(mapped_compartment_key) > 0:
        if len(mapped_compartment_key) > 1:
            logger.warning(
                f"More than one GO compartments for {comp_name} are mapped, using the first one"
            )

        compartment_entry = {
            SBML_DFS.C_ID: comp.getId(),
            SBML_DFS.C_NAME: comp.getName(),
            SBML_DFS.C_IDENTIFIERS: identifiers.Identifiers(
                [
                    identifiers.format_uri(
                        uri=identifiers.create_uri_url(
                            ontology=ONTOLOGIES.GO,
                            identifier=COMPARTMENTS_GO_TERMS[mapped_compartment_key[0]],
                        ),
                        biological_qualifier_type=BQB.IS,
                    )
                ]
            ),
            SBML_DFS.C_SOURCE: source.Source(init=True),
        }

    return compartment_entry


def _get_gene_product_dict(gp):
    """Extracts attributes of a gene product from an SBML reaction object.

    Parameters
    ----------
    gp : libsbml.GeneProduct
        A libsbml GeneProduct object.

    Returns
    -------
    dict
        A dictionary containing the gene product's ID, name, and identifiers.
    """
    return {
        SBML_DFS.RSC_ID: gp.getId(),
        SBML_DFS.SC_ID: gp.getGeneProduct(),
        SBML_DFS.STOICHIOMETRY: 0,
        SBML_DFS.SBO_TERM: gp.getSBOTermID(),
    }


def _extract_gene_products(association: libsbml.Association) -> list[dict]:
    """Recursively extracts gene products from an association tree."""
    gene_products = []

    def _recursive_helper(assoc: libsbml.Association):
        if hasattr(assoc, SBML_DEFS.REACTION_ATTR_GET_GENE_PRODUCT):
            gene_products.append(_get_gene_product_dict(assoc))
        elif hasattr(assoc, "getNumAssociations"):
            for i in range(assoc.getNumAssociations()):
                _recursive_helper(assoc.getAssociation(i))

    _recursive_helper(association)
    return gene_products



from __future__ import annotations

import gzip
import os
import tempfile
from datetime import datetime
from pathlib import Path
from unittest.mock import Mock, patch

import numpy as np
import pandas as pd
import pytest
from fs.tarfs import TarFS
from fs.zipfs import ZipFS
from google.cloud import storage
from pytest import fixture
from testcontainers.core.container import DockerContainer

from napistu import utils
from napistu.utils import drop_extra_cols


@fixture(scope="session")
def gcs_storage():
    """A container running a GCS emulator"""
    with (
        DockerContainer("fsouza/fake-gcs-server:1.44")
        .with_bind_ports(4443, 4443)
        .with_command("-scheme http -backend memory")
    ) as gcs:
        os.environ["STORAGE_EMULATOR_HOST"] = "http://0.0.0.0:4443"
        yield gcs


@fixture
def gcs_bucket_name(gcs_storage):
    bucket_name = f"testbucket-{datetime.now().strftime('%Y%m%d%H%M%S%f')}"
    return bucket_name


@fixture
def gcs_bucket(gcs_bucket_name):
    """A GCS bucket"""
    client = storage.Client()
    client.create_bucket(gcs_bucket_name)
    bucket = client.bucket(gcs_bucket_name)
    yield bucket
    bucket.delete(force=True)


@fixture
def gcs_bucket_uri(gcs_bucket, gcs_bucket_name):
    return f"gs://{gcs_bucket_name}"


@fixture
def gcs_bucket_subdir_uri(gcs_bucket_uri):
    return f"{gcs_bucket_uri}/testdir"


@fixture
def tmp_new_subdir(tmp_path):
    """An empty temporary directory"""
    return tmp_path / "test_dir"


def create_blob(bucket, blob_name, content=b"test"):
    # create the marker file
    bucket.blob(blob_name).upload_from_string(content)


def test_get_source_base_and_path_gcs():
    source_base, source_path = utils.get_source_base_and_path(
        "gs://cpr-ml-dev-us-east1/cpr/tests/test_data/pw_index.tsv"
    )
    assert source_base == "gs://cpr-ml-dev-us-east1"
    assert source_path == "cpr/tests/test_data/pw_index.tsv"


def test_get_source_base_and_path_local():
    source_base, source_path = utils.get_source_base_and_path(
        "/test_data/bla/pw_index.tsv"
    )
    assert source_base == "/test_data/bla"
    assert source_path == "pw_index.tsv"


def test_get_source_base_and_path_local_rel():
    source_base, source_path = utils.get_source_base_and_path(
        "./test_data/bla/pw_index.tsv"
    )
    assert source_base == "./test_data/bla"
    assert source_path == "pw_index.tsv"


def test_get_source_base_and_path_local_direct():
    source_base, source_path = utils.get_source_base_and_path("pw_index.tsv")
    assert source_base == ""
    assert source_path == "pw_index.tsv"


def test_initialize_dir_new(tmp_new_subdir):
    utils.initialize_dir(tmp_new_subdir, overwrite=False)
    assert tmp_new_subdir.exists()


@pytest.mark.unix_only
def test_initialize_dir_new_gcs(gcs_bucket_uri):
    test_uri = f"{gcs_bucket_uri}/testdir"
    utils.initialize_dir(test_uri, overwrite=False)
    utils.path_exists(test_uri)


def test_initialize_dir_new_2_layers(tmp_new_subdir):
    target_sub_dir = tmp_new_subdir / "test_dir_2"
    utils.initialize_dir(target_sub_dir, overwrite=False)
    assert target_sub_dir.exists()


@pytest.mark.unix_only
def test_initialize_dir_new_2_layers_gcs(gcs_bucket_uri):
    test_uri = f"{gcs_bucket_uri}/testdir/testdir2"
    utils.initialize_dir(test_uri, overwrite=False)
    utils.path_exists(test_uri)


def test_initialize_dir_existing(tmp_new_subdir):
    tmp_new_subdir.mkdir()

    test_file = tmp_new_subdir / "test_file"
    test_file.touch()

    with pytest.raises(FileExistsError):
        utils.initialize_dir(tmp_new_subdir, overwrite=False)
    assert test_file.exists()

    utils.initialize_dir(tmp_new_subdir, overwrite=True)
    assert test_file.exists() is False


@pytest.mark.unix_only
def test_initialize_dir_existing_gcs(gcs_bucket, gcs_bucket_uri):
    # create the file
    create_blob(gcs_bucket, "testdir/file")
    # This is a drawback of the current implementation - folders are only
    # recognized if they have a marker file.
    create_blob(gcs_bucket, "testdir/")

    test_uri = f"{gcs_bucket_uri}/testdir"
    test_uri_file = f"{test_uri}/file"
    with pytest.raises(FileExistsError):
        utils.initialize_dir(test_uri, overwrite=False)
        assert utils.path_exists(test_uri_file)

    utils.initialize_dir(test_uri, overwrite=True)
    assert utils.path_exists(test_uri_file) is False


def mock_targ_gz(url, tmp_file):
    with TarFS(tmp_file, write=True) as fol:
        with fol.open("test.txt", "w") as f:
            f.write("test")


def mock_zip(url, tmp_file):
    with ZipFS(tmp_file, write=True) as fol:
        with fol.open("test.txt", "w") as f:
            f.write("test")


def mock_gz(url, tmp_file):
    with gzip.open(tmp_file, mode="wt") as f:
        f.write("test")


@patch("napistu.utils.download_wget", side_effect=mock_targ_gz)
def test_download_and_extract_tar_gz(mock_download, tmp_new_subdir):
    utils.download_and_extract(
        url="http://asdf/bla.tar.gz",
        output_dir_path=tmp_new_subdir,
        download_method="wget",
    )
    assert (tmp_new_subdir / "test.txt").exists()


@patch("napistu.utils.download_ftp", side_effect=mock_zip)
def test_download_and_extract_zip(mock_download, tmp_new_subdir):
    utils.download_and_extract(
        url="http://asdf/bla.txt.zip",
        output_dir_path=tmp_new_subdir,
        download_method="ftp",
    )
    assert (tmp_new_subdir / "test.txt").exists()


@patch("napistu.utils.download_wget", side_effect=mock_gz)
def test_download_and_extract_gz(mock_download, tmp_new_subdir):
    utils.download_and_extract(
        url="http://asdf/bla.txt.gz",
        output_dir_path=tmp_new_subdir,
        download_method="wget",
    )
    assert (tmp_new_subdir / "bla.txt").exists()


def test_download_and_extract_invalid_method(tmp_new_subdir):
    with pytest.raises(ValueError):
        utils.download_and_extract(
            url="http://asdf/bla.txt.zip",
            output_dir_path=tmp_new_subdir,
            download_method="bla",
        )


@patch("napistu.utils.download_ftp", side_effect=mock_zip)
def test_download_and_extract_invalid_ext(mock_download, tmp_new_subdir):
    with pytest.raises(ValueError):
        utils.download_and_extract(
            url="http://asdf/bla.txt.zipper",
            output_dir_path=tmp_new_subdir,
            download_method="ftp",
        )


def test_path_exists(tmp_path, tmp_new_subdir):
    assert utils.path_exists(tmp_path)
    assert utils.path_exists(tmp_new_subdir) is False
    fn = tmp_path / "test.txt"
    assert utils.path_exists(fn) is False
    fn.touch()
    assert utils.path_exists(fn)
    assert utils.path_exists(".")
    tmp_new_subdir.mkdir()
    assert utils.path_exists(tmp_new_subdir)


@pytest.mark.unix_only
def test_path_exists_gcs(gcs_bucket, gcs_bucket_uri):
    assert utils.path_exists(gcs_bucket_uri)
    test_dir = "testdir"
    gcs_test_dir_uri = f"{gcs_bucket_uri}/{test_dir}"
    assert utils.path_exists(gcs_test_dir_uri) is False
    # Create the marker file for the directory, such that it 'exists'
    create_blob(gcs_bucket, f"{test_dir}/")
    assert utils.path_exists(gcs_test_dir_uri)

    # Test if files exists
    test_file = f"{test_dir}/test.txt"
    gcs_test_file_uri = f"{gcs_bucket_uri}/{test_file}"
    assert utils.path_exists(gcs_test_file_uri) is False
    # create the file
    create_blob(gcs_bucket, test_file)
    assert utils.path_exists(gcs_test_file_uri)


@pytest.mark.unix_only
def test_save_load_pickle_existing_folder(tmp_path):
    fn = tmp_path / "test.pkl"
    payload = "test"
    utils.save_pickle(fn, payload)
    assert fn.exists()
    assert utils.load_pickle(fn) == payload


@pytest.mark.skip_on_windows
def test_save_load_pickle_new_folder(tmp_new_subdir):
    fn = tmp_new_subdir / "test.pkl"
    payload = "test"
    utils.save_pickle(fn, payload)
    assert fn.exists()
    assert utils.load_pickle(fn) == payload


@pytest.mark.unix_only
def test_save_load_pickle_existing_folder_gcs(gcs_bucket_uri):
    fn = f"{gcs_bucket_uri}/test.pkl"
    payload = "test"
    utils.save_pickle(fn, payload)
    assert utils.path_exists(fn)
    assert utils.load_pickle(fn) == payload


@pytest.mark.unix_only
def test_save_load_pickle_new_folder_gcs(gcs_bucket_subdir_uri):
    fn = f"{gcs_bucket_subdir_uri}/test.pkl"
    payload = "test"
    utils.save_pickle(fn, payload)
    assert utils.path_exists(fn)
    assert utils.load_pickle(fn) == payload


@pytest.mark.skip_on_windows
def test_copy_uri_file(tmp_path, tmp_new_subdir):
    basename = "test.txt"
    fn = tmp_path / basename
    fn.write_text("test")
    fn_out = tmp_new_subdir / "test_out.txt"
    utils.copy_uri(fn, fn_out)
    assert fn_out.read_text() == "test"


@pytest.mark.skip_on_windows
def test_copy_uri_fol(tmp_path, tmp_new_subdir):
    tmp_new_subdir.mkdir()
    (tmp_new_subdir / "test").touch()

    out_dir = tmp_path / "out"
    out_file = out_dir / "test"
    utils.copy_uri(tmp_new_subdir, out_dir, is_file=False)
    assert out_file.exists()


@pytest.mark.unix_only
def test_copy_uri_file_gcs(gcs_bucket_uri, gcs_bucket_subdir_uri):
    basename = "test.txt"
    content = "test"
    fn = f"{gcs_bucket_uri}/{basename}"
    utils.save_pickle(fn, content)
    fn_out = f"{gcs_bucket_subdir_uri}/{basename}"
    utils.copy_uri(fn, fn_out)
    assert utils.path_exists(fn_out)
    assert utils.load_pickle(fn_out) == content


@pytest.mark.unix_only
def test_copy_uri_fol_gcs(gcs_bucket_uri, gcs_bucket_subdir_uri):
    basename = "test.txt"
    content = "test"
    fn = f"{gcs_bucket_subdir_uri}/{basename}"
    utils.save_pickle(fn, content)
    out_dir = f"{gcs_bucket_uri}/new_dir"
    out_file = f"{out_dir}/{basename}"
    utils.copy_uri(gcs_bucket_subdir_uri, out_dir, is_file=False)
    assert utils.path_exists(out_file)


@pytest.mark.skip_on_windows
def test_pickle_cache(tmp_path):
    fn = tmp_path / "test.pkl"

    mock = Mock()
    result = "test"

    @utils.pickle_cache(fn)
    def test_func():
        mock()
        return result

    test_func()
    r = test_func()
    assert r == result
    # only called once as second
    # call should be cached
    assert mock.call_count == 1


def test_extract_regex():
    assert utils.extract_regex_search("ENS[GT][0-9]+", "ENST0005") == "ENST0005"
    assert utils.extract_regex_search("ENS[GT]([0-9]+)", "ENST0005", 1) == "0005"
    with pytest.raises(ValueError):
        utils.extract_regex_search("ENS[GT][0-9]+", "ENSA0005")

    assert utils.extract_regex_match(".*type=([a-zA-Z]+).*", "Ltype=abcd5") == "abcd"
    # use for formatting identifiers
    assert utils.extract_regex_match("^([a-zA-Z]+)_id$", "sc_id") == "sc"
    with pytest.raises(ValueError):
        utils.extract_regex_match(".*type=[a-zA-Z]+.*", "Ltype=abcd5")


def test_match_pd_vars():
    a_series = pd.Series({"foo": 1, "bar": 2})
    a_dataframe = pd.DataFrame({"foo": ["a", "b"], "bar": [1, 2]})

    assert utils.match_pd_vars(a_series, {"foo", "bar"}).are_present
    assert not utils.match_pd_vars(a_series, {"baz"}).are_present
    assert utils.match_pd_vars(a_dataframe, {"foo", "bar"}).are_present
    assert not utils.match_pd_vars(a_dataframe, {"baz"}).are_present


def test_ensure_pd_df():
    source_df = pd.DataFrame({"a": "b"}, index=[0])
    source_series = pd.Series({"a": "b"}).rename(0)

    converted_series = utils.ensure_pd_df(source_series)

    assert isinstance(utils.ensure_pd_df(source_df), pd.DataFrame)
    assert isinstance(converted_series, pd.DataFrame)
    assert all(converted_series.index == source_df.index)
    assert all(converted_series.columns == source_df.columns)
    assert all(converted_series == source_df)


def test_format_identifiers_as_edgelist():
    DEGEN_EDGELIST_DF_1 = pd.DataFrame(
        {
            "ind1": [0, 0, 1, 1, 1, 1],
            "ind2": ["a", "a", "b", "b", "c", "d"],
            "ont": ["X", "X", "X", "Y", "Y", "Y"],
            "val": ["A", "B", "C", "D", "D", "E"],
        }
    ).set_index(["ind1", "ind2"])

    DEGEN_EDGELIST_DF_2 = pd.DataFrame(
        {
            "ind": ["a", "a", "b", "b", "c", "d"],
            "ont": ["X", "X", "X", "Y", "Y", "Y"],
            "val": ["A", "B", "C", "D", "D", "E"],
        }
    ).set_index("ind")

    edgelist_df = utils.format_identifiers_as_edgelist(
        DEGEN_EDGELIST_DF_1, ["ont", "val"]
    )
    assert edgelist_df["ind"].iloc[0] == "ind_0_a"
    assert edgelist_df["id"].iloc[0] == "id_X_A"

    edgelist_df = utils.format_identifiers_as_edgelist(DEGEN_EDGELIST_DF_1, ["val"])
    assert edgelist_df["ind"].iloc[0] == "ind_0_a"
    assert edgelist_df["id"].iloc[0] == "id_A"

    edgelist_df = utils.format_identifiers_as_edgelist(
        DEGEN_EDGELIST_DF_2, ["ont", "val"]
    )
    assert edgelist_df["ind"].iloc[0] == "ind_a"
    assert edgelist_df["id"].iloc[0] == "id_X_A"

    with pytest.raises(ValueError):
        utils.format_identifiers_as_edgelist(
            DEGEN_EDGELIST_DF_2.reset_index(drop=True), ["ont", "val"]
        )


def test_find_weakly_connected_subgraphs():
    DEGEN_EDGELIST_DF_2 = pd.DataFrame(
        {
            "ind": ["a", "a", "b", "b", "c", "d"],
            "ont": ["X", "X", "X", "Y", "Y", "Y"],
            "val": ["A", "B", "C", "D", "D", "E"],
        }
    ).set_index("ind")

    edgelist_df = utils.format_identifiers_as_edgelist(
        DEGEN_EDGELIST_DF_2, ["ont", "val"]
    )
    edgelist = edgelist_df[["ind", "id"]]

    connected_indices = utils.find_weakly_connected_subgraphs(edgelist[["ind", "id"]])
    assert all(connected_indices["cluster"] == [0, 1, 1, 2])


def test_style_df():
    np.random.seed(0)
    simple_df = pd.DataFrame(np.random.randn(20, 4), columns=["A", "B", "C", "D"])
    simple_df.index.name = "foo"

    multiindexed_df = (
        pd.DataFrame(
            {
                "category": ["foo", "foo", "foo", "bar", "bar", "bar"],
                "severity": ["major", "minor", "minor", "major", "major", "minor"],
            }
        )
        .assign(message="stuff")
        .groupby(["category", "severity"])
        .count()
    )

    # style a few pd.DataFrames
    isinstance(utils.style_df(simple_df), pd.io.formats.style.Styler)
    isinstance(
        utils.style_df(simple_df, headers=None, hide_index=True),
        pd.io.formats.style.Styler,
    )
    isinstance(
        utils.style_df(simple_df, headers=["a", "b", "c", "d"], hide_index=True),
        pd.io.formats.style.Styler,
    )
    isinstance(utils.style_df(multiindexed_df), pd.io.formats.style.Styler)


def test_score_nameness():
    assert utils.score_nameness("p53") == 23
    assert utils.score_nameness("ENSG0000001") == 56
    assert utils.score_nameness("pyruvate kinase") == 15


def test_click_str_to_list():
    assert utils.click_str_to_list("['foo', bar]") == ["foo", "bar"]
    with pytest.raises(ValueError):
        utils.click_str_to_list("foo")


def test_drop_extra_cols():
    """Test the _drop_extra_cols function for removing and reordering columns."""
    # Setup test DataFrames
    df_in = pd.DataFrame({"col1": [1, 2, 3], "col2": [4, 5, 6], "col3": [7, 8, 9]})

    df_out = pd.DataFrame(
        {
            "col2": [10, 11, 12],
            "col3": [13, 14, 15],
            "col4": [16, 17, 18],  # Extra column that should be dropped
            "col1": [19, 20, 21],  # Different order than df_in
        }
    )

    # Test basic functionality without always_include
    result = drop_extra_cols(df_in, df_out)

    # Check that extra column was dropped
    assert "col4" not in result.columns

    # Check that columns are in the same order as df_in
    assert list(result.columns) == list(df_in.columns)

    # Check that values are preserved
    pd.testing.assert_frame_equal(
        result,
        pd.DataFrame(
            {"col1": [19, 20, 21], "col2": [10, 11, 12], "col3": [13, 14, 15]}
        )[
            list(df_in.columns)
        ],  # Ensure same column order
    )

    # Test with always_include
    result_with_include = drop_extra_cols(df_in, df_out, always_include=["col4"])

    # Check that col4 is retained and appears at the end
    assert list(result_with_include.columns) == list(df_in.columns) + ["col4"]
    assert result_with_include["col4"].equals(df_out["col4"])

    # Test with always_include containing non-existent column
    result_non_existent = drop_extra_cols(
        df_in, df_out, always_include=["col4", "col5"]
    )
    assert list(result_non_existent.columns) == list(df_in.columns) + ["col4"]

    # Test with always_include containing column from df_in
    result_overlap = drop_extra_cols(df_in, df_out, always_include=["col1", "col4"])
    assert list(result_overlap.columns) == list(df_in.columns) + ["col4"]

    # Test with no overlapping columns but some in always_include
    df_out_no_overlap = pd.DataFrame({"col4": [1, 2, 3], "col5": [4, 5, 6]})
    result_no_overlap = drop_extra_cols(df_in, df_out_no_overlap)
    assert result_no_overlap.empty
    assert list(result_no_overlap.columns) == []

    result_no_overlap_with_include = drop_extra_cols(
        df_in, df_out_no_overlap, always_include=["col4"]
    )
    assert list(result_no_overlap_with_include.columns) == ["col4"]
    assert result_no_overlap_with_include["col4"].equals(df_out_no_overlap["col4"])

    # Test with subset of columns
    df_out_subset = pd.DataFrame(
        {"col1": [1, 2, 3], "col3": [7, 8, 9], "col4": [10, 11, 12]}
    )
    result_subset = drop_extra_cols(df_in, df_out_subset)

    assert list(result_subset.columns) == ["col1", "col3"]
    pd.testing.assert_frame_equal(result_subset, df_out_subset[["col1", "col3"]])

    result_subset_with_include = drop_extra_cols(
        df_in, df_out_subset, always_include=["col4"]
    )
    assert list(result_subset_with_include.columns) == ["col1", "col3", "col4"]
    pd.testing.assert_frame_equal(
        result_subset_with_include, df_out_subset[["col1", "col3", "col4"]]
    )


def test_merge_and_log_overwrites(caplog):
    """Test merge_and_log_overwrites function."""

    # Test basic merge with no conflicts
    df1 = pd.DataFrame({"id": [1, 2], "value1": ["a", "b"]})
    df2 = pd.DataFrame({"id": [1, 2], "value2": ["c", "d"]})
    result = utils._merge_and_log_overwrites(df1, df2, "test", on="id")
    assert set(result.columns) == {"id", "value1", "value2"}
    assert len(caplog.records) == 0

    # Test merge with column conflict
    df1 = pd.DataFrame({"id": [1, 2], "name": ["a", "b"], "value": [10, 20]})
    df2 = pd.DataFrame({"id": [1, 2], "name": ["c", "d"]})
    result = utils._merge_and_log_overwrites(df1, df2, "test", on="id")

    # Check that the right columns exist
    assert set(result.columns) == {"id", "name", "value"}
    # Check that we got df2's values for the overlapping column
    assert list(result["name"]) == ["c", "d"]
    # Check that we kept df1's non-overlapping column
    assert list(result["value"]) == [10, 20]
    # Check that the warning was logged
    assert len(caplog.records) == 1
    assert "test merge" in caplog.records[0].message
    assert "name" in caplog.records[0].message

    # Test merge with multiple column conflicts
    caplog.clear()
    df1 = pd.DataFrame(
        {"id": [1, 2], "name": ["a", "b"], "value": [10, 20], "status": ["ok", "ok"]}
    )
    df2 = pd.DataFrame(
        {"id": [1, 2], "name": ["c", "d"], "status": ["pending", "done"]}
    )
    result = utils._merge_and_log_overwrites(df1, df2, "test", on="id")

    # Check that the right columns exist
    assert set(result.columns) == {"id", "name", "value", "status"}
    # Check that we got df2's values for the overlapping columns
    assert list(result["name"]) == ["c", "d"]
    assert list(result["status"]) == ["pending", "done"]
    # Check that we kept df1's non-overlapping column
    assert list(result["value"]) == [10, 20]
    # Check that the warning was logged with both column names
    assert len(caplog.records) == 1
    assert "test merge" in caplog.records[0].message
    assert "name" in caplog.records[0].message
    assert "status" in caplog.records[0].message

    # Test merge with index
    caplog.clear()
    df1 = pd.DataFrame({"name": ["a", "b"], "value": [10, 20]}, index=[1, 2])
    df2 = pd.DataFrame({"name": ["c", "d"]}, index=[1, 2])
    result = utils._merge_and_log_overwrites(
        df1, df2, "test", left_index=True, right_index=True
    )

    # Check that the right columns exist
    assert set(result.columns) == {"name", "value"}
    # Check that we got df2's values for the overlapping column
    assert list(result["name"]) == ["c", "d"]
    # Check that we kept df1's non-overlapping column
    assert list(result["value"]) == [10, 20]
    # Check that the warning was logged
    assert len(caplog.records) == 1
    assert "test merge" in caplog.records[0].message
    assert "name" in caplog.records[0].message


def test_matrix_to_edgelist():
    # Test case 1: Basic functionality with numeric indices
    matrix = np.array([[1, 2, np.nan], [np.nan, 3, 4], [5, np.nan, 6]])
    expected_edgelist = pd.DataFrame(
        {
            "row": [0, 0, 1, 1, 2, 2],
            "column": [0, 1, 1, 2, 0, 2],
            "value": np.array([1, 2, 3, 4, 5, 6], dtype=np.float64),
        }
    )
    result = utils.matrix_to_edgelist(matrix)
    pd.testing.assert_frame_equal(result, expected_edgelist)

    # Test case 2: With row and column labels
    row_labels = ["A", "B", "C"]
    col_labels = ["X", "Y", "Z"]
    expected_labeled_edgelist = pd.DataFrame(
        {
            "row": ["A", "A", "B", "B", "C", "C"],
            "column": ["X", "Y", "Y", "Z", "X", "Z"],
            "value": np.array([1, 2, 3, 4, 5, 6], dtype=np.float64),
        }
    )
    result_labeled = utils.matrix_to_edgelist(
        matrix, row_labels=row_labels, col_labels=col_labels
    )
    pd.testing.assert_frame_equal(result_labeled, expected_labeled_edgelist)

    # Test case 3: Empty matrix (all NaN)
    empty_matrix = np.full((2, 2), np.nan)
    empty_result = utils.matrix_to_edgelist(empty_matrix)
    assert empty_result.empty

    # Test case 4: Single value matrix
    single_matrix = np.array([[1]])
    expected_single = pd.DataFrame(
        {"row": [0], "column": [0], "value": np.array([1], dtype=np.int64)}
    )
    result_single = utils.matrix_to_edgelist(single_matrix)
    pd.testing.assert_frame_equal(result_single, expected_single)


def test_safe_fill():
    safe_fill_test = ["a_very_long stringggg", ""]
    assert [utils.safe_fill(x) for x in safe_fill_test] == [
        "a_very_long\nstringggg",
        "",
    ]


def test_update_pathological_names():

    # All numeric
    s = pd.Series(["1", "2", "3"])
    out = utils.update_pathological_names(s, "prefix_")
    assert all(x.startswith("prefix_") for x in out)
    assert list(out) == ["prefix_1", "prefix_2", "prefix_3"]

    # Mixed numeric and non-numeric
    s2 = pd.Series(["1", "foo", "3"])
    out2 = utils.update_pathological_names(s2, "prefix_")
    assert list(out2) == ["1", "foo", "3"]

    # All non-numeric
    s3 = pd.Series(["foo", "bar", "baz"])
    out3 = utils.update_pathological_names(s3, "prefix_")
    assert list(out3) == ["foo", "bar", "baz"]


def test_parquet_save_load():
    """Test that write_parquet and read_parquet work correctly."""
    # Create test DataFrame
    original_df = pd.DataFrame(
        {
            "sc_id_origin": ["A", "B", "C"],
            "sc_id_dest": ["B", "C", "A"],
            "path_length": [1, 2, 3],
            "path_weights": [0.1, 0.5, 0.8],
            "has_connection": [True, False, True],
        }
    )

    # Write and read using temporary file
    with tempfile.TemporaryDirectory() as temp_dir:
        file_path = Path(temp_dir) / "test.parquet"
        utils.save_parquet(original_df, file_path)
        result_df = utils.load_parquet(file_path)

        # Verify they're identical
        pd.testing.assert_frame_equal(original_df, result_df)



"""
Utilities for loading and processing documentation.
"""

import httpx
import os
from napistu.constants import PACKAGE_DEFS
from napistu.mcp.constants import (
    DEFAULT_GITHUB_API,
    GITHUB_ISSUES_INDEXED,
    GITHUB_PRS_INDEXED,
)


async def load_readme_content(readme_url: str) -> str:
    if readme_url.startswith("http://") or readme_url.startswith("https://"):
        async with httpx.AsyncClient() as client:
            response = await client.get(readme_url)
            response.raise_for_status()
            return response.text
    else:
        raise ValueError(
            f"Only HTTP(S) URLs are supported for documentation paths: {readme_url}"
        )


async def fetch_wiki_page(
    page_name: str,
    repo: str = PACKAGE_DEFS.GITHUB_PROJECT_REPO,
    owner: str = PACKAGE_DEFS.GITHUB_OWNER,
) -> str:
    """
    Fetch wiki page content using raw GitHub URLs.

    Parameters
    ----------
    page_name : str
        The name of the page (without .md extension for wiki pages)
    repo : str
        The repository name
    owner : str
        The GitHub username or organization

    Returns
    -------
    str
        The raw Markdown content
    """
    # Use raw.githubusercontent.com for wiki pages
    url = f"https://raw.githubusercontent.com/wiki/{owner}/{repo}/{page_name}.md"

    async with httpx.AsyncClient() as client:
        resp = await client.get(url)
        resp.raise_for_status()
        return resp.text


async def list_issues(
    repo: str,
    owner: str = PACKAGE_DEFS.GITHUB_OWNER,
    github_api: str = DEFAULT_GITHUB_API,
    state: str = GITHUB_ISSUES_INDEXED,
    include_prs: bool = False,
) -> list:
    """
    List issues (and optionally PRs) for a given GitHub repository using the GitHub API.

    Parameters
    ----------
    repo : str, optional
        The repository name.
    owner : str, optional
        The GitHub username or organization (default is 'napistu').
    github_api : str, optional
        The GitHub API base URL (default is 'https://api.github.com').
    state : str, optional
        The state of the issues to return. Can be 'open', 'closed', or 'all'. Default is 'open'.
    include_prs : bool, optional
        If True, include pull requests in the results. Default is False.

    Returns
    -------
    list of dict
        Each dict contains: number, title, state, url, and a truncated body (max 500 chars).
    """
    url = f"{github_api}/repos/{owner}/{repo}/issues?state={state}"
    filter_func = (
        (lambda item: True)
        if include_prs
        else (lambda item: "pull_request" not in item)
    )
    return await _fetch_github_items(url, filter_func=filter_func)


async def list_pull_requests(
    repo: str,
    owner: str = PACKAGE_DEFS.GITHUB_OWNER,
    github_api: str = DEFAULT_GITHUB_API,
    state: str = GITHUB_PRS_INDEXED,
) -> list:
    """
    List pull requests for a given GitHub repository using the GitHub API.

    Parameters
    ----------
    repo : str, optional
        The repository name.
    owner : str, optional
        The GitHub username or organization (default is 'napistu').
    github_api : str, optional
        The GitHub API base URL (default is 'https://api.github.com').
    state : str, optional
        The state of the PRs to return. Can be 'open', 'closed', or 'all'. Default is 'open'.

    Returns
    -------
    list of dict
        Each dict contains: number, title, state, url, and a truncated body (max 500 chars).
    """
    url = f"{github_api}/repos/{owner}/{repo}/pulls?state={state}"
    return await _fetch_github_items(url)


async def get_issue(
    repo: str,
    number: int,
    owner: str = PACKAGE_DEFS.GITHUB_OWNER,
    github_api: str = DEFAULT_GITHUB_API,
) -> dict:
    """
    Get a single issue (or PR) by number from a GitHub repository.

    Parameters
    ----------
    repo : str
        The repository name.
    number : int
        The issue or PR number.
    owner : str, optional
        The GitHub username or organization (default is 'napistu').
    github_api : str, optional
        The GitHub API base URL (default is 'https://api.github.com').

    Returns
    -------
    dict
        The issue or PR details as a dictionary.
    """
    url = f"{github_api}/repos/{owner}/{repo}/issues/{number}"
    async with httpx.AsyncClient(headers=_get_github_headers()) as client:
        resp = await client.get(url)
        resp.raise_for_status()
        item = resp.json()
        return _format_github_issue(item)


def _get_github_headers():
    """
    Return headers for GitHub API requests, including Authorization if GITHUB_TOKEN is set.
    """
    headers = {}

    GITHUB_TOKEN = os.getenv("GITHUB_TOKEN")
    if GITHUB_TOKEN:
        print("Using token from environment variable GITHUB_TOKEN")
        headers["Authorization"] = f"token {GITHUB_TOKEN}"
    return headers


def _format_github_issue(item):
    """
    Format a GitHub issue or PR item into a standard dict.
    """
    return {
        "number": item["number"],
        "title": item["title"],
        "state": item["state"],
        "url": item["html_url"],
        "body": (
            (item["body"][:500] + "...")
            if item.get("body") and len(item["body"]) > 500
            else item.get("body")
        ),
        "is_pr": "pull_request" in item or "merged_at" in item,
    }


async def _fetch_github_items(url, filter_func=None):
    """
    Fetch and format a list of GitHub issues or PRs from a given API endpoint.
    """
    async with httpx.AsyncClient(headers=_get_github_headers()) as client:
        resp = await client.get(url)
        resp.raise_for_status()
        items = []
        for item in resp.json():
            if filter_func and not filter_func(item):
                continue
            items.append(_format_github_issue(item))
        return items



from __future__ import annotations

from importlib.metadata import PackageNotFoundError
from importlib.metadata import version

try:
    __version__ = version("napistu-py")
except PackageNotFoundError:
    # package is not installed
    pass



from __future__ import annotations

import os

import pandas as pd
import pytest
from napistu import consensus
from napistu import identifiers
from napistu import indices
from napistu import source
from napistu import sbml_dfs_core
from napistu.ingestion import sbml
from napistu.modify import pathwayannot
from napistu.constants import SBML_DFS, SBML_DFS_SCHEMA, SCHEMA_DEFS, IDENTIFIERS, BQB

test_path = os.path.abspath(os.path.join(__file__, os.pardir))
test_data = os.path.join(test_path, "test_data")


def test_reduce_to_consensus_ids():
    sbml_path = os.path.join(test_data, "R-HSA-1237044.sbml")

    # test aggregating by IDs, by moving from compartmentalized_species -> species

    sbml_model = sbml.SBML(sbml_path)
    comp_species_df = sbml_model._define_cspecies()
    comp_species_df.index.names = [SBML_DFS.S_ID]
    consensus_species, species_lookup = consensus.reduce_to_consensus_ids(
        comp_species_df,
        {
            SCHEMA_DEFS.PK: SBML_DFS.S_ID,
            SCHEMA_DEFS.ID: SBML_DFS.S_IDENTIFIERS,
            SCHEMA_DEFS.TABLE: SBML_DFS.SPECIES,
        },
    )

    assert isinstance(consensus_species, pd.DataFrame)
    assert consensus_species.shape == (18, 4)
    assert isinstance(species_lookup, pd.Series)
    assert species_lookup.size == 23


def test_consensus():
    pw_index = indices.PWIndex(os.path.join(test_data, "pw_index.tsv"))
    sbml_dfs_dict = consensus.construct_sbml_dfs_dict(pw_index)

    consensus_model = consensus.construct_consensus_model(sbml_dfs_dict, pw_index)
    assert consensus_model.species.shape == (38, 3)
    assert consensus_model.reactions.shape == (30, 4)
    assert consensus_model.reaction_species.shape == (137, 4)

    consensus_model = pathwayannot.drop_cofactors(consensus_model)
    assert consensus_model.species.shape == (38, 3)
    assert consensus_model.reaction_species.shape == (52, 4)
    # update reaction_species.shape after more cofactors identified

    consensus_model.validate()


def test_source_tracking():
    # create input data
    table_schema = {"source": "source_var", "pk": "primary_key"}

    # define existing sources and the new_id entity they belong to
    # here, we are assuming that each model has a blank source object
    # as if it came from a non-consensus model
    agg_tbl = pd.DataFrame(
        {
            "new_id": [0, 0, 1, 1],
        }
    )
    agg_tbl[table_schema["source"]] = source.Source(init=True)

    # define new_ids and the models they came from
    # these models will be matched to the pw_index to flush out metadata
    lookup_table = pd.DataFrame(
        {
            "new_id": [0, 0, 1, 1],
            "model": ["R-HSA-1237044", "R-HSA-425381", "R-HSA-1237044", "R-HSA-425381"],
        }
    )

    # use an existing pw_index since pw_index currently checks for the existence of the source file
    pw_index = indices.PWIndex(os.path.join(test_data, "pw_index.tsv"))

    # test create source table
    source_table = source.create_source_table(lookup_table, table_schema, pw_index)
    assert source_table["source_var"][0].source.shape == (2, 8)

    # test create_consensus_sources
    consensus_sources = consensus.create_consensus_sources(
        agg_tbl, lookup_table, table_schema, pw_index
    )
    assert consensus_sources[0].source.shape == (2, 8)

    # lets add a model which does not have a reference in the pw_index
    invalid_lookup_table = pd.DataFrame(
        {
            "new_id": [0, 0, 1, 1],
            "model": ["R-HSA-1237044", "R-HSA-425381", "R-HSA-1237044", "typo"],
        }
    )

    # expect a ValueError when the model is not found
    with pytest.raises(ValueError) as _:
        source.create_source_table(invalid_lookup_table, table_schema, pw_index)

    # now we will aggregate the consensus model above with a new single model (which has some
    # overlapping entries with the consensusd (id 1) and some new ids (id 2)

    agg_tbl2 = pd.DataFrame(
        {
            "new_id": [0, 1, 1, 2],
        }
    )

    agg_tbl2[table_schema["source"]] = consensus_sources.tolist() + [
        source.Source(init=True) for i in range(0, 2)
    ]

    lookup_table2 = pd.DataFrame(
        {
            "new_id": [0, 1, 1, 2],
            # the model for the first two entries should really correspond to the "consensus"
            # but since this is not a file I will stub with one of the pw_index entries
            "model": [
                "R-HSA-1247673",
                "R-HSA-1247673",
                "R-HSA-1475029",
                "R-HSA-1475029",
            ],
        }
    )

    source_table = source.create_source_table(lookup_table2, table_schema, pw_index)
    assert source_table.shape == (3, 1)
    assert [
        source_table["source_var"][i].source.shape
        for i in range(0, source_table.shape[0])
    ] == [(1, 8), (2, 8), (1, 8)]

    consensus_sources = consensus.create_consensus_sources(
        agg_tbl2, lookup_table2, table_schema, pw_index
    )
    assert [
        consensus_sources[i].source.shape for i in range(0, consensus_sources.shape[0])
    ] == [(3, 8), (4, 8), (1, 8)]


def test_passing_entity_data():

    pw_index = indices.PWIndex(os.path.join(test_data, "pw_index.tsv"))
    sbml_dfs_dict = consensus.construct_sbml_dfs_dict(pw_index)

    for model in list(sbml_dfs_dict.keys())[0:3]:
        sbml_dfs_dict[model].add_species_data(
            "my_species_data",
            sbml_dfs_dict[model]
            .species.iloc[0:5]
            .assign(my_species_data_var="testing")["my_species_data_var"]
            .to_frame(),
        )
        sbml_dfs_dict[model].add_reactions_data(
            "my_reactions_data",
            sbml_dfs_dict[model]
            .reactions.iloc[0:5]
            .assign(my_reactions_data_var1="testing")
            .assign(my_reactions_data_var2="testing2")[
                ["my_reactions_data_var1", "my_reactions_data_var2"]
            ],
        )

    # create a consensus with perfect merges of overlapping id-table-variable values
    # i.e., when combined all merged entries have the same attributes
    consensus_model = consensus.construct_consensus_model(sbml_dfs_dict, pw_index)

    assert len(consensus_model.species_data) == 1
    assert consensus_model.species_data["my_species_data"].shape == (10, 1)
    assert len(consensus_model.reactions_data) == 1
    assert consensus_model.reactions_data["my_reactions_data"].shape == (14, 2)

    # add different tables from different models
    for model in list(sbml_dfs_dict.keys())[3:5]:
        sbml_dfs_dict[model].add_species_data(
            "my_other_species_data",
            sbml_dfs_dict[model]
            .species.iloc[0:5]
            .assign(my_species_data="testing")["my_species_data"]
            .to_frame(),
        )

    consensus_model = consensus.construct_consensus_model(sbml_dfs_dict, pw_index)
    assert len(consensus_model.species_data) == 2

    # create a case where reactions will be merged and the same reaction
    # in different models has a different value for its reactions_data
    minimal_pw_index = pw_index
    minimal_pw_index.index = minimal_pw_index.index.iloc[0:2]

    # Since we're working with a DataFrame, we can use loc to update the file value directly
    minimal_pw_index.index.loc[1, "file"] = minimal_pw_index.index.loc[0, "file"]

    duplicated_sbml_dfs_dict = consensus.construct_sbml_dfs_dict(minimal_pw_index)
    # explicitely define the order we'll loop through models so that
    # the position of a model can be used to set mismatching attributes
    # for otherwise identical models
    model_order = list(duplicated_sbml_dfs_dict.keys())

    for model in duplicated_sbml_dfs_dict.keys():
        model_index = model_order.index(model)

        duplicated_sbml_dfs_dict[model].add_reactions_data(
            "my_mismatched_data",
            duplicated_sbml_dfs_dict[model]
            .reactions.iloc[0:5]
            .assign(my_reactions_data_var1=model)["my_reactions_data_var1"]
            .to_frame()
            .assign(numeric_var=[x + model_index for x in range(0, 5)])
            .assign(bool_var=[x + model_index % 2 == 0 for x in range(0, 5)]),
        )

    # assign reversibility is True for one model to
    # confirm that reversibility trumps irreversible
    # when merging reactions with identical stoichiometry but
    # different reversibility attributes

    duplicated_sbml_dfs_dict["R-HSA-1237044"].reactions = duplicated_sbml_dfs_dict[
        "R-HSA-1237044"
    ].reactions.assign(r_isreversible=True)

    consensus_model = consensus.construct_consensus_model(
        duplicated_sbml_dfs_dict, pw_index
    )
    assert consensus_model.reactions_data["my_mismatched_data"].shape == (5, 3)
    assert consensus_model.reactions["r_isreversible"].eq(True).all()


def test_consensus_ontology_check():
    pw_index = indices.PWIndex(os.path.join(test_data, "pw_index.tsv"))

    test_sbml_dfs_dict = consensus.construct_sbml_dfs_dict(pw_index)
    test_consensus_model = consensus.construct_consensus_model(
        test_sbml_dfs_dict, pw_index
    )

    pre_shared_onto_sp_list, pre_onto_df = consensus.pre_consensus_ontology_check(
        test_sbml_dfs_dict, "species"
    )
    assert set(pre_shared_onto_sp_list) == {"chebi", "reactome", "uniprot"}

    post_shared_onto_sp_set = consensus.post_consensus_species_ontology_check(
        test_consensus_model
    )
    assert post_shared_onto_sp_set == {"chebi", "reactome", "uniprot"}


def test_report_consensus_merges_reactions(tmp_path):
    # Create two minimal SBML_dfs objects with a single reaction each, same r_id
    r_id = "R00000001"
    reactions = pd.DataFrame(
        {
            SBML_DFS.R_NAME: ["rxn1"],
            SBML_DFS.R_IDENTIFIERS: [None],
            SBML_DFS.R_SOURCE: [None],
            SBML_DFS.R_ISREVERSIBLE: [False],
        },
        index=[r_id],
    )
    reactions.index.name = SBML_DFS.R_ID
    reaction_species = pd.DataFrame(
        {
            SBML_DFS.R_ID: [r_id],
            SBML_DFS.SC_ID: ["SC0001"],
            SBML_DFS.STOICHIOMETRY: [1],
            SBML_DFS.SBO_TERM: ["SBO:0000459"],
        },
        index=["RSC0001"],
    )
    reaction_species.index.name = SBML_DFS.RSC_ID
    compartmentalized_species = pd.DataFrame(
        {
            SBML_DFS.SC_NAME: ["A [cytosol]"],
            SBML_DFS.S_ID: ["S0001"],
            SBML_DFS.C_ID: ["C0001"],
            SBML_DFS.SC_SOURCE: [None],
        },
        index=["SC0001"],
    )
    compartmentalized_species.index.name = SBML_DFS.SC_ID
    species = pd.DataFrame(
        {
            SBML_DFS.S_NAME: ["A"],
            SBML_DFS.S_IDENTIFIERS: [None],
            SBML_DFS.S_SOURCE: [None],
        },
        index=["S0001"],
    )
    species.index.name = SBML_DFS.S_ID
    compartments = pd.DataFrame(
        {
            SBML_DFS.C_NAME: ["cytosol"],
            SBML_DFS.C_IDENTIFIERS: [None],
            SBML_DFS.C_SOURCE: [None],
        },
        index=["C0001"],
    )
    compartments.index.name = SBML_DFS.C_ID
    sbml_dict = {
        SBML_DFS.COMPARTMENTS: compartments,
        SBML_DFS.SPECIES: species,
        SBML_DFS.COMPARTMENTALIZED_SPECIES: compartmentalized_species,
        SBML_DFS.REACTIONS: reactions,
        SBML_DFS.REACTION_SPECIES: reaction_species,
    }
    sbml1 = sbml_dfs_core.SBML_dfs(sbml_dict, validate=False, resolve=False)
    sbml2 = sbml_dfs_core.SBML_dfs(sbml_dict, validate=False, resolve=False)
    sbml_dfs_dict = {"mod1": sbml1, "mod2": sbml2}

    # Create a lookup_table that merges both reactions into a new_id
    lookup_table = pd.DataFrame(
        {
            "model": ["mod1", "mod2"],
            "r_id": [r_id, r_id],
            "new_id": ["merged_rid", "merged_rid"],
        }
    )
    # Use the reactions schema
    table_schema = SBML_DFS_SCHEMA.SCHEMA[SBML_DFS.REACTIONS]

    # Call the function and check that it runs and the merge_labels are as expected
    consensus.report_consensus_merges(
        lookup_table.set_index(["model", "r_id"])[
            "new_id"
        ],  # this is a Series with name 'new_id'
        table_schema,
        sbml_dfs_dict=sbml_dfs_dict,
        n_example_merges=1,
    )
    # No assertion: this is a smoke test to ensure the Series output is handled without error


def test_build_consensus_identifiers_handles_merges_and_missing_ids():

    # Three entities:
    # - 'A' with identifier X
    # - 'B' with no identifiers
    # - 'C' with identifier X (should merge with 'A')
    df = pd.DataFrame(
        {
            "s_id": ["A", "B", "C"],
            "s_Identifiers": [
                identifiers.Identifiers(
                    [
                        {
                            IDENTIFIERS.ONTOLOGY: "test",
                            IDENTIFIERS.IDENTIFIER: "X",
                            IDENTIFIERS.BQB: BQB.IS,
                        }
                    ]
                ),
                identifiers.Identifiers([]),
                identifiers.Identifiers(
                    [
                        {
                            IDENTIFIERS.ONTOLOGY: "test",
                            IDENTIFIERS.IDENTIFIER: "X",
                            IDENTIFIERS.BQB: BQB.IS,
                        }
                    ]
                ),
            ],
        }
    ).set_index("s_id")

    schema = SBML_DFS_SCHEMA.SCHEMA[SBML_DFS.SPECIES]

    indexed_cluster, cluster_consensus_identifiers = (
        consensus.build_consensus_identifiers(df, schema)
    )

    # All entities should be assigned to a cluster
    assert set(indexed_cluster.index) == set(df.index)
    assert not indexed_cluster.isnull().any()
    # There should be a consensus identifier for each cluster
    assert set(cluster_consensus_identifiers.index) == set(indexed_cluster.values)

    # Entities 'A' and 'C' should be merged (same cluster)
    assert indexed_cluster.loc["A"] == indexed_cluster.loc["C"]
    # Entity 'B' should be in a different cluster
    assert indexed_cluster.loc["B"] != indexed_cluster.loc["A"]

    # The consensus identifier for the merged cluster should include identifier X
    merged_cluster_id = indexed_cluster.loc["A"]
    ids_obj = cluster_consensus_identifiers.loc[merged_cluster_id, schema["id"]]
    assert any(i["identifier"] == "X" for i in getattr(ids_obj, "ids", []))

    # The consensus identifier for the entity with no identifiers should be empty
    noid_cluster_id = indexed_cluster.loc["B"]
    ids_obj_noid = cluster_consensus_identifiers.loc[noid_cluster_id, schema["id"]]
    assert hasattr(ids_obj_noid, "ids")
    assert len(getattr(ids_obj_noid, "ids", [])) == 0


################################################
# __main__
################################################

if __name__ == "__main__":
    test_reduce_to_consensus_ids()
    test_consensus()
    test_source_tracking()
    test_passing_entity_data()
    test_consensus_ontology_check()
    test_report_consensus_merges_reactions()
    test_build_consensus_identifiers_handles_merges_and_missing_ids()



# src/napistu/mcp/health.py
"""
Health check endpoint for the MCP server when deployed to Cloud Run.
"""

import logging
from typing import Dict, Any, TypeVar
from datetime import datetime
from fastmcp import FastMCP

logger = logging.getLogger(__name__)

# Type variable for the FastMCP decorator return type
T = TypeVar("T")

# Global cache for component health status
_health_cache = {"status": "initializing", "components": {}, "last_check": None}


def register_components(mcp: FastMCP) -> None:
    """
    Register health check components with the MCP server.

    Parameters
    ----------
    mcp : FastMCP
        FastMCP server instance to register the health endpoint with.
    """

    @mcp.resource("napistu://health")
    async def health_check() -> Dict[str, Any]:
        """
        Health check endpoint for deployment monitoring.
        Returns current cached health status.
        """
        return _health_cache

    @mcp.tool()
    async def check_health() -> Dict[str, Any]:
        """
        Tool to actively check current component health.
        This performs real-time checks and updates the cached status.
        """
        global _health_cache
        try:
            health_status = {
                "status": "healthy",
                "timestamp": datetime.utcnow().isoformat(),
                "version": _get_version(),
                "components": await _check_components(),
            }

            # Check if any components failed
            failed_components = [
                name
                for name, status in health_status["components"].items()
                if status["status"] == "unavailable"
            ]

            if failed_components:
                health_status["status"] = "degraded"
                health_status["failed_components"] = failed_components

            # Update the global cache with latest status
            health_status["last_check"] = datetime.utcnow().isoformat()
            _health_cache.update(health_status)
            logger.info(f"Updated health cache - Status: {health_status['status']}")

            return health_status

        except Exception as e:
            logger.error(f"Health check failed: {e}")
            error_status = {
                "status": "unhealthy",
                "error": str(e),
                "timestamp": datetime.utcnow().isoformat(),
                "last_check": datetime.utcnow().isoformat(),
            }
            # Update cache even on error
            _health_cache.update(error_status)
            return error_status


async def initialize_components() -> bool:
    """
    Initialize health check components.
    Performs initial health check and caches the result.

    Returns
    -------
    bool
        True if initialization is successful
    """
    global _health_cache

    logger.info("Initializing health check components...")

    try:
        # Check initial component health
        component_status = await _check_components()

        # Update cache
        _health_cache.update(
            {
                "status": "healthy",
                "components": component_status,
                "timestamp": datetime.utcnow().isoformat(),
                "version": _get_version(),
                "last_check": datetime.utcnow().isoformat(),
            }
        )

        # Check for failed components
        failed_components = [
            name
            for name, status in component_status.items()
            if status["status"] == "unavailable"
        ]

        if failed_components:
            _health_cache["status"] = "degraded"
            _health_cache["failed_components"] = failed_components

        logger.info(f"Health check initialization complete: {_health_cache['status']}")
        return True

    except Exception as e:
        logger.error(f"Health check initialization failed: {e}")
        _health_cache["status"] = "unhealthy"
        _health_cache["error"] = str(e)
        return False


def _check_component_health(component_name: str, module_path: str) -> Dict[str, Any]:
    """
    Check the health of a single MCP component using the component class pattern.

    Parameters
    ----------
    component_name : str
        Name of the component (for logging)
    module_path : str
        Full module path for importing the component

    Returns
    -------
    Dict[str, Any]
        Dictionary containing component health status from the component's state
    """
    try:
        # Import the component module
        module = __import__(module_path, fromlist=[component_name])

        # Use the new component class pattern
        if hasattr(module, "get_component"):
            try:
                component = module.get_component()
                state = component.get_state()
                health_status = state.get_health_status()
                logger.info(f"{component_name} health: {health_status}")
                return health_status
            except RuntimeError as e:
                # Handle execution component that might not be created yet
                if "not created" in str(e):
                    logger.warning(f"{component_name} not initialized yet")
                    return {
                        "status": "initializing",
                        "message": "Component not created",
                    }
                else:
                    raise
        else:
            # Component doesn't follow the new pattern
            logger.warning(f"{component_name} doesn't use component class pattern")
            return {"status": "unknown", "message": "Component using legacy pattern"}

    except ImportError as e:
        logger.error(f"Could not import {component_name}: {str(e)}")
        return {"status": "unavailable", "error": f"Import failed: {str(e)}"}
    except Exception as e:
        logger.error(f"{component_name} health check failed: {str(e)}")
        return {"status": "unavailable", "error": str(e)}


async def _check_components() -> Dict[str, Dict[str, Any]]:
    """
    Check the health of individual MCP components using their component classes.

    Returns
    -------
    Dict[str, Dict[str, Any]]
        Dictionary mapping component names to their health status
    """
    # Define component configurations
    component_configs = {
        "documentation": "napistu.mcp.documentation",
        "codebase": "napistu.mcp.codebase",
        "tutorials": "napistu.mcp.tutorials",
        "execution": "napistu.mcp.execution",
    }

    logger.info("Starting component health checks...")
    logger.info(f"Checking components: {list(component_configs.keys())}")

    # Check each component using their state objects
    results = {
        name: _check_component_health(name, module_path)
        for name, module_path in component_configs.items()
    }

    logger.info(f"Health check results: {results}")
    return results


def _get_version() -> str:
    """
    Get the Napistu version.

    Returns
    -------
    str
        Version string of the Napistu package, or 'unknown' if not available.
    """
    try:
        import napistu

        return getattr(napistu, "__version__", "unknown")
    except ImportError:
        return "unknown"



"""
Core MCP server implementation for Napistu.
"""

import asyncio
import logging

from mcp.server import FastMCP

from napistu.mcp import codebase
from napistu.mcp import documentation
from napistu.mcp import execution
from napistu.mcp import tutorials
from napistu.mcp import health

from napistu.mcp.profiles import ServerProfile, get_profile
from napistu.mcp.constants import MCP_DEFAULTS
from napistu.mcp.config import MCPServerConfig

logger = logging.getLogger(__name__)


def _register_component(
    name: str, module, config_key: str, config: dict, mcp: FastMCP, **kwargs
) -> None:
    """
    Register a single component with the MCP server.

    Parameters
    ----------
    name : str
        Component name for logging
    module : module
        Component module with get_component() function or create_component() for execution
    config_key : str
        Configuration key to check if component is enabled
    config : dict
        Server configuration
    mcp : FastMCP
        FastMCP server instance
    **kwargs : dict
        Additional arguments for component creation (used by execution component)
    """
    if not config.get(config_key, False):
        return  # Skip disabled components

    logger.info(f"Registering {name} components")

    if name == "execution":
        # Special handling for execution component which needs session context
        component = module.create_component(
            session_context=kwargs.get("session_context"),
            object_registry=kwargs.get("object_registry"),
        )
    else:
        component = module.get_component()

    component.register(mcp)


def create_server(profile: ServerProfile, server_config: MCPServerConfig) -> FastMCP:
    """
    Create an MCP server based on a profile configuration and server config.

    Parameters
    ----------
    profile : ServerProfile
        Server profile to use. All configuration must be set in the profile. (Valid profiles: 'execution', 'docs', 'full')
    server_config : MCPServerConfig
        Server configuration with validated host, port, and server name.

    Returns
    -------
    FastMCP
        Configured FastMCP server instance.
    """

    config = profile.get_config()

    # Create the server with validated configuration
    mcp = FastMCP(
        server_config.server_name, host=server_config.host, port=server_config.port
    )

    # Define component configurations
    component_configs = [
        ("documentation", documentation, "enable_documentation"),
        ("codebase", codebase, "enable_codebase"),
        ("tutorials", tutorials, "enable_tutorials"),
        ("execution", execution, "enable_execution"),
    ]

    # Register all components
    for name, module, config_key in component_configs:
        _register_component(
            name,
            module,
            config_key,
            config,
            mcp,
            session_context=config.get("session_context"),
            object_registry=config.get("object_registry"),
        )

    # Always register health components
    health.register_components(mcp)
    logger.info("Registered health components")

    return mcp


async def _initialize_component(
    name: str, module, config_key: str, config: dict
) -> bool:
    """
    Initialize a single component with error handling.

    Parameters
    ----------
    name : str
        Component name for logging
    module : module
        Component module with get_component() function
    config_key : str
        Configuration key to check if component is enabled
    config : dict
        Server configuration

    Returns
    -------
    bool
        True if initialization successful
    """
    if not config.get(config_key, False):
        return True  # Skip disabled components

    logger.info(f"Initializing {name} components")
    try:
        component = module.get_component()
        result = await component.safe_initialize()
        return result
    except Exception as e:
        logger.error(f"❌ {name.title()} components failed to initialize: {e}")
        return False


async def initialize_components(profile: ServerProfile) -> None:
    """
    Asynchronously initialize all enabled components for the MCP server.

    Parameters
    ----------
    profile : ServerProfile
        The profile whose configuration determines which components to initialize.

    Returns
    -------
    None
    """
    config = profile.get_config()

    # Define component configurations
    component_configs = [
        ("documentation", documentation, "enable_documentation"),
        ("codebase", codebase, "enable_codebase"),
        ("tutorials", tutorials, "enable_tutorials"),
        ("execution", execution, "enable_execution"),
    ]

    # Initialize all components
    initialization_results = {}

    for name, module, config_key in component_configs:
        result = await _initialize_component(name, module, config_key, config)
        initialization_results[name] = result

    # Initialize health components last since they monitor the other components
    logger.info("Initializing health components")
    try:
        result = await health.initialize_components()
        initialization_results["health"] = result
        if result:
            logger.info("✅ Health components initialized successfully")
        else:
            logger.warning("⚠️ Health components initialized with issues")
    except Exception as e:
        logger.error(f"❌ Health components failed to initialize: {e}")
        initialization_results["health"] = False

    # Summary of initialization
    successful = sum(1 for success in initialization_results.values() if success)
    total = len(initialization_results)
    logger.info(
        f"Component initialization complete: {successful}/{total} components successful"
    )

    if successful == 0:
        logger.error(
            "❌ All components failed to initialize - server may not function correctly"
        )
    elif successful < total:
        logger.warning(
            "⚠️ Some components failed to initialize - server running in degraded mode"
        )


def start_mcp_server(profile_name: str, server_config: MCPServerConfig) -> None:
    """
    Start MCP server - main entry point for server startup.

    Parameters
    ----------
    profile_name : str
        Server profile to use ('local', 'remote', 'full').
    server_config : MCPServerConfig
        Validated server configuration.

    Returns
    -------
    None
        This function runs indefinitely until interrupted.

    Notes
    -----
    The server uses HTTP transport (streamable-http) for all connections.
    Components are initialized asynchronously before the server starts.
    Health components are always registered and initialized last.
    """
    # Set up logging
    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger("napistu")

    logger.info("Starting Napistu MCP Server")
    logger.info(f"  Profile: {profile_name}")
    logger.info(f"  Host: {server_config.host}")
    logger.info(f"  Port: {server_config.port}")
    logger.info(f"  Server Name: {server_config.server_name}")
    logger.info(f"  Transport: {MCP_DEFAULTS.TRANSPORT}")

    # Create session context for execution components
    session_context = {}
    object_registry = {}

    # Get profile with configuration
    profile: ServerProfile = get_profile(
        profile_name,
        session_context=session_context,
        object_registry=object_registry,
        server_name=server_config.server_name,
    )

    # Create server with validated configuration
    mcp = create_server(profile, server_config)

    # Initialize components first (separate async call)
    async def init_components():
        logger.info("Initializing MCP components...")
        await initialize_components(profile)
        logger.info("✅ Component initialization complete")

    # Run initialization
    asyncio.run(init_components())

    # Debug info
    logger.info(f"Server settings: {mcp.settings}")

    logger.info("🚀 Starting MCP server...")
    logger.info(
        f"Using {MCP_DEFAULTS.TRANSPORT} transport on http://{server_config.host}:{server_config.port}"
    )

    mcp.run(transport=MCP_DEFAULTS.TRANSPORT)
